from pathlib import Path

import yaml
from yaml.loader import SafeLoader

class YAMLFile():
    """
    YAMLFile is base class that is used to read and write YAML files
    """
    def __init__(self, path: Path):
        super().__init__()
        self._path = path

    @property
    def path(self) -> Path:
        return self._path
    
    def exists(self) -> bool:
        return self._path.exists()
    
    def read(self) -> dict:
        with open(self._path, 'r') as stream:
            return yaml.load(stream, Loader=SafeLoader)
        
    def __str__(self) -> str:
        return self._path.as_posix()
    

class PureMLConfigYML():
    """
    PureMLConfigYML is a class that is used to generate the configuration file `puremlconfig.yaml`
    It is capable of reading and writing the configuration file
    """
    def __init__(self, path: Path):
        self._yaml_file = YAMLFile(path)
        self._yaml_document = None 

    @property
    def file(self) -> YAMLFile:
        return self._yaml_file
    
    @property
    def data(self) -> dict:
        if self._yaml_document is None:
            if self._yaml_file.exists():
                self._yaml_document = self._yaml_file.read()
            else:
                self._yaml_document = {}
        return self._yaml_document
    
    def save(self, data: dict = None) -> None:
        if data is None:
            data = self.data
        if len(data.keys()) == 0:
            print("No configuration data to save")
            return
        with open(self._yaml_file.path, 'w') as stream:
            yaml.dump(data, stream, default_flow_style=False)
    
    def reload(self) -> None:
        self._yaml_document = None


from pydantic import model_validator, ConfigDict, BaseModel, validator
from .singleton import Singleton_BaseModel
import typing
import os
from typing import ClassVar

# project_path = os.getcwd()
# if os.path.exists(project_path / "puremlconfig.yaml"):
#     puremlconfig = PureMLConfigYML(project_path / "puremlconfig.yaml")
# else:
#     puremlconfig = None

class PathSchema(Singleton_BaseModel):

    # Probably for future relative path configuration using config file
    # if puremlconfig is not None:
    #     if str(puremlconfig.data["repository"]).startswith("file://"):
    #         relative_base_url = str(puremlconfig.data["repository"]).removeprefix("file://")

    PATH_PUREML_RELATIVE:ClassVar[str] = ".pureml"
    PATH_PREDICT_DIR_RELATIVE:ClassVar[str] = "predict"

    PATH_USER_TOKEN:ClassVar[str] = os.path.join(
        os.path.expanduser("~"), PATH_PUREML_RELATIVE, "token"
    )

    PATH_USER_PROJECT_DIR:ClassVar[str] = os.path.join(os.getcwd(), PATH_PUREML_RELATIVE)

    PATH_USER_PROJECT :ClassVar[str] = os.path.join(PATH_USER_PROJECT_DIR, "pure.project")

    PATH_CONFIG:ClassVar[str] = os.path.join(PATH_USER_PROJECT_DIR, "config.pkl")  # 'temp.yaml'

    PATH_ARTIFACT_DIR:ClassVar[str] = os.path.join(PATH_USER_PROJECT_DIR, "artifacts")
    PATH_ARRAY_DIR:ClassVar[str] = os.path.join(PATH_USER_PROJECT_DIR, "array")
    PATH_AUDIO_DIR:ClassVar[str] = os.path.join(PATH_USER_PROJECT_DIR, "audio")
    PATH_FIGURE_DIR:ClassVar[str] = os.path.join(PATH_USER_PROJECT_DIR, "figure")
    PATH_TABULAR_DIR:ClassVar[str] = os.path.join(PATH_USER_PROJECT_DIR, "tabular")
    PATH_VIDEO_DIR:ClassVar[str] = os.path.join(PATH_USER_PROJECT_DIR, "video")
    PATH_IMAGE_DIR:ClassVar[str] = os.path.join(PATH_USER_PROJECT_DIR, "image")

    PATH_DATASET_DIR:ClassVar[str] = os.path.join(PATH_USER_PROJECT_DIR, "dataset")

    PATH_MODEL_DIR:ClassVar[str] = os.path.join(PATH_USER_PROJECT_DIR, "model")

    PATH_PREDICT_DIR:ClassVar[str] = os.path.join(PATH_USER_PROJECT_DIR, PATH_PREDICT_DIR_RELATIVE)
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode = 'after')
    @classmethod
    def create_base_folders(cls, values):
        os.makedirs(values.PATH_USER_PROJECT_DIR, exist_ok=True)

        return values

    @model_validator(mode = 'after')
    @classmethod
    def create_log_folders(cls, values):
        os.makedirs(values.PATH_ARTIFACT_DIR, exist_ok=True)
        os.makedirs(values.PATH_ARRAY_DIR, exist_ok=True)
        os.makedirs(values.PATH_AUDIO_DIR, exist_ok=True)
        os.makedirs(values.PATH_FIGURE_DIR, exist_ok=True)
        os.makedirs(values.PATH_TABULAR_DIR, exist_ok=True)
        os.makedirs(values.PATH_VIDEO_DIR, exist_ok=True)
        os.makedirs(values.PATH_IMAGE_DIR, exist_ok=True)

        return values

    @model_validator(mode = 'after')
    @classmethod
    def create_model_folders(cls, values):
        os.makedirs(values.PATH_MODEL_DIR, exist_ok=True)

        return values

    @model_validator(mode = 'after')
    @classmethod
    def create_dataset_folders(cls,values):
        os.makedirs(values.PATH_DATASET_DIR, exist_ok=True)

        return values

    @model_validator(mode = 'after')
    @classmethod
    def create_predict_folders(cls, values):
        os.makedirs(values.PATH_PREDICT_DIR, exist_ok=True)

        return values



import requests
from rich import print

import os
import json

from pureml.cli.helpers import get_auth_headers
from pureml.storage import upload_and_get_provider_and_path

from . import get_org_id
from pureml.schema import (
    ModelSchema,
    StorageSchema,
    ConfigKeys,
    ContentTypeHeader,
    AcceptHeader,
)
from pureml import save_model, load_model
from urllib.parse import urljoin
import joblib
from pureml.utils.hash import generate_hash_for_file
from pureml.utils.readme import load_readme
from pureml.utils.version_utils import parse_version_label
from pureml.utils.config import reset_config
from pureml.utils.logger import get_logger


config_keys = ConfigKeys
storage = StorageSchema().get_instance()
logger = get_logger("sdk.components.model")


def check_model_hash(hash: str, label: str):
    logger.info("Checking model hash")

    name, _ = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    url = "org/{}/model/{}/hash-status".format(org_id, name)
    url = urljoin(model_schema._backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    data = {"hash": hash}

    data = json.dumps(data)

    response = requests.post(url, data=data, headers=headers)

    hash_exists = False

    if response.ok:
        hash_exists = response.json()["data"][0]

    logger.info(f"Hash exists: {hash_exists}")
    return hash_exists


def list():
    """This function will return a list of all the modelst

    Returns
    -------
        A list of all the models

    """
    logger.info("Fetching list of models")

    org_id = get_org_id()
    model_schema = ModelSchema()

    url = "org/{}/model/all".format(org_id)
    url = urljoin(model_schema._backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        # print(f"[green]Obtained list of models")

        response_text = response.json()
        model_list = response_text["data"]
        # print(model_list)

        return model_list
    else:
        logger.error("Unable to obtain the list of models", response=response)
        print(f"[orange]Unable to obtain the list of models!")

    return


def init(label: str, type: str = "ML", readme: str = None):
    logger.info("Initializing model")
    name, _ = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    if readme is None:
        readme = ModelSchema().PATH_MODEL_README

    file_content, file_type = load_readme(path=readme)

    url = "org/{}/model/{}/create".format(org_id, name)
    url = urljoin(model_schema._backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_JSON)

    data = {
        "name": name,
        "readme": {"file_type": file_type, "content": file_content},
        "type": type,
    }

    data = json.dumps(data)

    # files = {"file": (readme, open(readme, "rb"), file_type)}

    response = requests.post(url, data=data, headers=headers)
    # response = requests.post(url, data=data, headers=headers, files=files)

    if response.ok:
        print(f"[green]Model has been created!")

        return True
    else:
        logger.error("Unable to create the model", response=response)
        print(f"[orange]Model has not been created!")

        return False


def register(
    model,
    label,
    is_empty: bool = False,
):
    logger.info("Registering model")
    name, _ = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    model_file_name = ".".join([name, "pkl"])
    model_path = os.path.join(model_schema._paths.PATH_MODEL_DIR, model_file_name)

    os.makedirs(model_schema._paths.PATH_MODEL_DIR, exist_ok=True)

    save_model(model, name, model_path=model_path)

    model_hash = generate_hash_for_file(
        file_path=model_path, name=name, is_empty=is_empty
    )

    model_exists = model_status(label)

    if not model_exists:
        model_created = init(label)
        # print("model_created", model_created)
        if not model_created:
            print("[orange] Unable to register the model")
            return False, model_hash, "latest"

    model_exists_remote = check_model_hash(hash=model_hash, label=label)

    if model_exists_remote:
        print(f"[green]Model already exists. Not registering a new version![/green]")
        return True, model_hash, "latest"
    else:
        url = "org/{}/model/{}/register".format(org_id, name)
        url = urljoin(model_schema._backend.BASE_URL, url)

        headers = get_auth_headers(content_type=None, accept=AcceptHeader.APP_JSON)

        # files = {"file": (model_file_name, open(model_path, "rb"))}
        provider, file_path = upload_and_get_provider_and_path(
            model_path, opt_base_dir="models"
        )

        data = {
            "name": name,
            "hash": model_hash,
            "is_empty": is_empty,
            "storage": provider,
            "file_path": file_path,
        }

        response = requests.post(url, files={"foo": "bar"}, data=data, headers=headers)

        if response.ok:

            model_version = None

            # print(response.json())
            try:
                print(f"[green]Model has been registered!")

                model_version = response.json()["data"][0]["version"]
                print("Model version: ", model_version)
                model_label = ":".join([name, model_version])
                print("Model label: ", model_label)

                # reset_config(key=config_keys.model.value)
            except Exception as e:
                print("[orange] Incorrect json response. Model has not been registered")
                print(e)

            return True, model_hash, model_version

        else:
            logger.error("Unable to register the model", response=response)
            print(f"[orange]Model has not been registered!")
            print(response.text)

        return False, model_hash, None


def model_status(label: str):
    logger.info("Checking model status")
    name, _ = parse_version_label(label)

    model_details = details(label=label)

    logger.info(f"Model details: {model_details}")
    if model_details:
        return True
    else:
        return False


def details(label: str):
    """It fetches the details of a model.

    Parameters
    ----------
    name : str
        The name of the model
    version: str
        The version of the model
    Returns
    -------
        The details of the model.

    """
    logger.info("Fetching model details")

    name, _ = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    url = "org/{}/model/{}".format(org_id, name)
    url = urljoin(model_schema._backend.BASE_URL, url)

    headers = get_auth_headers(
        content_type=ContentTypeHeader.APP_FORM_URL_ENCODED,
        accept=AcceptHeader.APP_JSON,
    )

    response = requests.get(url, headers=headers)
    # print(response.url)
    # print(response.text)

    if response.ok:
        # print(f"[green]Model details have been fetched")
        response_text = response.json()
        model_details = response_text["data"][0]
        # print(model_details)

        return model_details

    else:
        logger.error("Unable to fetch model details", response=response)
        print(f"[orange]Model details have not been found")
        return


def version_details(label: str):
    """It fetches the details of a model.

    Parameters
    ----------
    name : str
        The name of the model
    version: str
        The version of the model
    Returns
    -------
        The details of the model.

    """
    logger.info("Fetching model version details")

    name, version = parse_version_label(label)

    org_id = get_org_id()
    model_schema = ModelSchema()

    url = "org/{}/model/{}/version/{}".format(org_id, name, version)
    url = urljoin(model_schema._backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        # print(f"[green]Model Version details have been fetched")
        response_text = response.json()
        model_details = response_text["data"][0]
        # print(model_details)

        return model_details

    else:
        logger.error("Unable to fetch model version details", response=response)
        print(f"[orange]Model details have not been found")
        return


def fetch(label: str):
    """This function fetches a model from the server and returns it as a `Model` object

    Parameters
    ----------
    name : str, optional
        The name of the model you want to fetch.
    version: str
        The version of the model

    Returns
    -------
        The model is being returned.

    """
    logger.info("Fetching model")

    name, version = parse_version_label(label)

    org_id = get_org_id()

    model_details = version_details(label=label)

    if model_details is None:
        print(f"[orange]Unable to fetch Model version")
        return

    is_empty = model_details["is_empty"]

    if is_empty:
        print("[orange]Model file is not registered to the version")
        return

    # storage_path = model_details["path"]["source_path"]
    # storage_source_type = model_details["path"]["source_type"]["public_url"]

    # model_url = urljoin(storage_source_type, storage_path)

    model_url = model_details["path"]
    source_type = model_details["source_type"]

    logger.info(f"Model URL: {model_url}")
    logger.info(f"Source type: {source_type}")
    if source_type != "file":
        response = requests.get(model_url)

        if response.ok:
            model_bytes = response.content
            open("temp_model.pure", "wb").write(model_bytes)

            model = load_model(model_path="temp_model.pure")

            # print(f"[green]Model version has been fetched")
            return model
        else:
            logger.error("Unable to fetch Model version", response=response)
            print(f"[orange]Unable to fetch Model version")
            # print(response.status_code)
            # print(response.text)
            # print(response.url)
            return
    else:
        if os.path.exists(model_url):
            model = load_model(model_path=model_url)

            return model
        else:
            logger.error("Unable to fetch Model from local path", path=model_url)
            print(f"[orange]Unable to fetch Model")
            return


# def delete(label: str) -> str:
#     """This function deletes a model from the project

#     Parameters
#     ----------
#     name : str
#         The name of the model you want to delete
#     version : str
#         The version of the model to delete.

#     """

#     name, _ = parse_version_label(label)

#     user_token = get_token()
#     org_id = get_org_id()
#     model_schema = ModelSchema()

#     url = "org/{}/model/{}/delete".format(org_id, name)
#     url = urljoin(model_schema.backend.BASE_URL, url)

#     headers = {
#         "Content-Type": ContentTypeHeader.APP_FORM_URL_ENCODED.value,
#         "Authorization": "Bearer {}".format(user_token),
#     }

#     response = requests.delete(url, headers=headers)

#     if response.ok:
#         print(f"[green]Model has been deleted")

#     else:
#         print(f"[orange]Unable to delete Model")

#     return response.text


def serve_model():
    pass



from pathlib import Path
import typing

from ..model_framework import ModelFramework, ModelFrameworkType
from ..packaging_utils import infer_requirements



class SKLearn(ModelFramework):
    framework_name: str = 'scikit-learn'
    additional_requirements: list = [
                                        "numpy",
                                        "scipy",
                                        "joblib"
                                    ]
    requirements: list = [framework_name]



    def typ(self) -> ModelFrameworkType:
        return ModelFrameworkType.SKLEARN

    # def infer_requirements(self) -> typing.Dict[str, str]:
    #     pass
        # return infer_sklearn_packages()

    # def serialize_model_to_directory(self, model, target_directory: Path):
    #     import joblib

    #     model_filename = MODEL_FILENAME
    #     model_filepath = target_directory / model_filename
    #     joblib.dump(model, model_filepath, compress=True)

    # def model_metadata(self, model) -> Dict[str, str]:
    #     supports_predict_proba = model_supports_predict_proba(model)
    #     return {
    #         "model_binary_dir": "model",
    #         "supports_predict_proba": supports_predict_proba,
    #     }



    def supports_model_class(self, model_class) -> bool:
        model_framework, _, _ = model_class.__module__.partition(".")
        return model_framework == ModelFrameworkType.SKLEARN.value


    def get_requirements(self):

        default_requirements = [infer_requirements(framework_name=self.framework_name)]

        other_requirements = [infer_requirements(lib_name) for lib_name in self.additional_requirements]

        self.requirements = default_requirements + other_requirements


        # self.requirements = self.requirements + self.additional_requirements
        # print(self.requirements)
        # self.requirements  = [infer_requirements(lib_name) for lib_name in self.requirements]

        return self.requirements


    def load_model(self, model_config_path:str= None):
        pass
        # model_config = ModelConfig().load_from_disk(model_path=self.model_path)

    def predict(self):
        pass


from pydantic import ConfigDict, BaseModel, root_validator

from pureml.schema.singleton import Singleton_BaseModel


class StorageSchema(Singleton_BaseModel):

    STORAGE: str = "PUREML-STORAGE"
    model_config = ConfigDict(arbitrary_types_allowed=True)



from pydantic import ConfigDict, BaseModel
from pureml.predictor.predictor import BasePredictor
from pureml.components import dataset
from typing import Any
from importlib import import_module
from rich import print
import requests
from pureml_evaluate.evaluators.evaluator import eval as eval_fn
from pureml.cli.auth import get_auth_headers
from pureml.components import get_org_id


class EvalHelper(BaseModel):  # To pass the requirements to eval in pureml_evaluate
    task_type: None = None
    label_model: str
    label_dataset: str
    predictor: BasePredictor = None
    predictor_path: str = "predict.py"
    dataset: Any = None
    model_config = ConfigDict(validate_assignment=True, arbitrary_types_allowed=True)

    def load_dataset(self):
        # print(self.label_dataset)
        self.dataset = dataset.fetch(self.label_dataset)
        print("[green] Succesfully fetched the dataset")

    def load_predictor(self):
        module_path = self.predictor_path.replace(".py", "")
        module_import = import_module(module_path)

        predictor_class = getattr(module_import, "Predictor")

        self.predictor = predictor_class()
        print("[green] Succesfully fetched the predictor")

    def load_model(self):
        self.predictor.load_models()
        print("[green] Succesfully fetched the model")

    def load(self):
        self.load_dataset()
        self.load_predictor()
        self.load_model()

    def get_y_pred(self):
        return self.predictor.predict(self.dataset["x_test"])

    def get_y_true(self):
        return self.dataset["y_test"]

    def get_sensitive_features(self):
        # print(f"Dataset Keys: {self.dataset.keys()}")
        if 'sensitive_features' in self.dataset.keys():
            # print(f"Data in sensitive_features: {self.dataset['sensitive_features']}")
            return self.dataset['sensitive_features']
        # else:
        #    return None

    # This Function is to Format Risk Name. Eg - balanced_acc_score to Balanced Acc Score
    def format_riskname(self, metric_name):
        metricname = metric_name.split('_')
        return (
            metricname[0].capitalize()
            if len(metricname) == 1
            else ' '.join(word.capitalize() for word in metricname))

    def post(self, values):
        base_url = "https://pureml-development.up.railway.app/api"
        risk_url = "/org/{orgId}/model/{modelName}/version/{version}/risk"
        orgID = get_org_id()
        model_details = self.label_model.split(":")
        model_name = model_details[0]
        version = model_details[1]
        complete_url = base_url + \
            risk_url.format(orgId=orgID, modelName=model_name, version=version)
        # print(complete_url)
        response_status = 200
        if 'performance' in values['complete']:
            category = "performance"
            metric_name = values['complete']['performance']
            for i in metric_name:
                risk = self.format_riskname(i)
                severity = values['complete']['performance'][i]['severity']
                value = values['complete']['performance'][i]['value']
                value = format(value, '.2f')
                description = values['complete']['performance'][i]['description']
                matters = values['complete']['performance'][i]['matters']
                summary = values['complete']['performance'][i]['summary']
                threshold = values['complete']['performance'][i]['threshold']
                payload = {
                    "category": f"{category}",
                    "description": f"{description}",
                    "matters": f"{matters}",
                    "risk": f"{risk}",
                    "severity": f"{severity}",
                    "summary": f"{summary}",
                    "threshold": str(threshold),
                    "value": str(value)
                }
                # print(payload)
                try:
                    response = requests.post(
                        complete_url, json=payload, headers=get_auth_headers())
                    print(f"Response Status Code: {response.status_code}")
                    response_status = response.status_code
                except Exception as e:
                    print(e)

        if 'fairness' in values['complete']:
            category = "fairness"
            metric_name = values['complete']['fairness']
            for i in metric_name:
                risk = self.format_riskname(i)
                severity = values['complete']['fairness'][i]['severity']
                value = values['complete']['fairness'][i]['value']
                value = format(value, '.2f')
                description = values['complete']['fairness'][i]['description']
                matters = values['complete']['fairness'][i]['matters']
                summary = values['complete']['fairness'][i]['summary']
                threshold = values['complete']['fairness'][i]['threshold']
                payload = {
                    "category": f"{category}",
                    "description": f"{description}",
                    "matters": f"{matters}",
                    "risk": f"{risk}",
                    "severity": f"{severity}",
                    "summary": f"{summary}",
                    "threshold": str(threshold),
                    "value": str(value)
                }
                # print(payload)
                try:
                    response = requests.post(
                        complete_url, json=payload, headers=get_auth_headers())
                    print(f"Status Code: {response.status_code}")
                    response_status = response.status_code
                except Exception as e:
                    print(e)

        response = requests.get(complete_url, headers=get_auth_headers())
        print(response.json())
        if response_status == 200:
            print(
                "[green] Succesfully sent the evaluation results to PureML server")


def eval(label_model: str, label_dataset: str, task_type: list[str], metrics=None, path_to_config=None, pdf_file_name='metrics_graph.pdf'):
    evaluator = EvalHelper(
        task_type=None, label_dataset=label_dataset, label_model=label_model
    )

    evaluator.load()

    y_pred = evaluator.get_y_pred()
    y_true = evaluator.get_y_true()
    sensitive_features = evaluator.get_sensitive_features()
    # sensitive_features=None

    values = eval_fn(y_true=y_true, y_pred=y_pred, sensitive_features=sensitive_features,
                     evaluators=task_type, y_pred_scores=None, metrics=metrics,
                     path_to_config=path_to_config, as_pdf=True, pdf_file_name=pdf_file_name)

    # print()
    # evaluator.post(values)

    # print(data)
    # print(values)
    # result_split_json(values=values, label_model=label_model)
    # return values


def result_split_json(values, label_model):
    performance = values['complete']['performance']
    fairness = values['complete']['fairness']

    # severity_json = {}
    # for i in performance:
    #     severity_json[i] = performance[i]['severity']

    # values_json = {}
    # for i in performance:
    #     values_json[i] = performance[i]['value']

    for i in performance:
        payload = {
            "category": "performance",
            "risk": i,
            "severity": performance[i]['severity'],
            "value": str(performance[i]['value']),
            "threshold": str(performance[i]['threshold']),
            "policy": "nyc144"
        }
    base_url = "https://pureml-development.up.railway.app/api"
    risk_url = "/org/{orgId}/model/{modelName}/version/{version}/risk"
    orgID = get_org_id()
    model_details = label_model.split(":")
    model_name = model_details[0]
    version = model_details[1]
    complete_url = base_url + \
        risk_url.format(orgId=orgID, modelName=model_name, version=version)
    # print(complete_url)
    response_status = 200
    # print(payload)
    try:
        response = requests.post(
            complete_url, json=payload, headers=get_auth_headers())
        print(f"Status Code: {response.status_code}")
        # print(f"Reason: {response.reason}")
        # print(response.json())
        response_status = response.status_code
    except Exception as e:
        print(e)

    response = requests.get(complete_url, headers=get_auth_headers())
    print(response.json())
    if response_status == 200:
        print("[green] Succesfully sent the evaluation results to PureML server")

    for i in fairness:
        payload = {
            "category": "fairness",
            "risk": i,
            "severity": fairness[i]['severity'],
            "value": str(fairness[i]['value']),
            "threshold": str(fairness[i]['threshold']),
            "policy": "nyc144"

        }
    base_url = "https://pureml-development.up.railway.app/api"
    risk_url = "/org/{orgId}/model/{modelName}/version/{version}/risk"
    orgID = get_org_id()
    model_details = label_model.split(":")
    model_name = model_details[0]
    version = model_details[1]
    complete_url = base_url + \
        risk_url.format(orgId=orgID, modelName=model_name, version=version)
    # print(complete_url)
    response_status = 200
    # print(payload)
    try:
        response = requests.post(
            complete_url, json=payload, headers=get_auth_headers())
        print(f"Status Code: {response.status_code}")
        # print(f"Reason: {response.reason}")
        # print(response.json())
        response_status = response.status_code
    except Exception as e:
        print(e)

    response = requests.get(complete_url, headers=get_auth_headers())
    print(response.json())
    if response_status == 200:
        print("[green] Succesfully sent the evaluation results to PureML server")



from pureml.utils.pipeline import add_load_data_to_config


def requirements(requirements:str|list[str], **kwargs):

    def decorator(func):
        
        def wrapper(*args, **kwargs):
            
            func_output = func(*args, **kwargs)

            try:
                add_load_data_to_config(name=name, func=func)
            except Exception as e:
                print('Unable to add requirements to config')
                print(e)

            return func_output

        # print("Outside  wrapper")

        return wrapper
    # print('Outside decorator')
        
    return decorator


from .config import load_config, save_config
from .hash import generate_hash_for_dict, generate_hash_for_function
from .source_code import get_source_code
import os
import shutil
from .log_utils import update_step_dict
import time
from pureml.schema import ConfigKeys
from datetime import datetime

config_keys = ConfigKeys


def add_load_data_to_config(name, description=None, func=None, hash=""):

    config = load_config()
    code = ""
    if func is not None:
        try:
            code = get_source_code(func)
            hash = generate_hash_for_function(func)
        except Exception as e:
            print("Unable to get load_data source code")
            print(e)

    current_datetime = datetime.now()
    current_datetime = str(current_datetime.strftime("%d %b %Y, %H:%M"))

    config[config_keys.load_data.value] = {
        "name": name,
        "desc": str(description),
        "time": current_datetime,
        "hash": hash,
        "type": config_keys.load_data.value,
        "code": code,
    }

    save_config(config=config)


def add_transformer_to_config(name, description=None, func=None, hash="", parent=None):

    config = load_config()
    # print(config)
    position = len(config[config_keys.transformer.value]) + 1

    if parent is None:
        if position == 1:

            if len(config[config_keys.load_data.value]) != 0:
                parent = config[config_keys.load_data.value]["name"]

        else:
            transformer_previous = config[config_keys.transformer.value][position - 1]
            parent = transformer_previous["name"]

    code = ""
    if func is not None:
        try:
            code = get_source_code(func)
            hash = generate_hash_for_function(func)
        except Exception as e:
            print("Unable to get transformer source code")
            print(e)

    current_datetime = datetime.now()
    current_datetime = str(current_datetime.strftime("%d %b %Y, %H:%M"))

    config[config_keys.transformer.value][position] = {
        "name": name,
        "desc": str(description),
        "time": current_datetime,
        "hash": hash,
        "type": config_keys.transformer.value,
        "parent": parent,
        "code": code,
    }
    # print('saveing configuration for ', name)
    save_config(config=config)


def add_dataset_to_config(
    name, description=None, func=None, hash="", version="", parent=None
):

    config = load_config()

    if parent is None:

        if len(config[config_keys.transformer.value]) != 0:
            config_transformer = config[config_keys.transformer.value]
            transformer_last = list(config_transformer.values())[-1]
            parent = transformer_last["name"]

    code = ""
    if func is not None:
        try:
            code = get_source_code(func)
            hash = generate_hash_for_function(func)
        except Exception as e:
            print("Unable to get dataset source code")
            print(e)

    current_datetime = datetime.now()
    current_datetime = str(current_datetime.strftime("%d %b %Y, %H:%M"))

    config[config_keys.dataset.value] = {
        "name": name,
        "desc": str(description),
        "time": current_datetime,
        "hash": hash,
        "type": config_keys.dataset.value,
        "version": version,
        "parent": parent,
        "code": code,
    }

    save_config(config=config)


def add_model_to_config(name, description=None, func=None, hash="", version=""):
    # name = ''
    # hash = ''
    # version = ''

    config = load_config()

    code = ""
    if func is not None:
        try:
            code = get_source_code(func)
            hash = generate_hash_for_function(func)
        except Exception as e:
            print("Unable to get model source code")
            print(e)

    current_datetime = datetime.now()
    current_datetime = str(current_datetime.strftime("%d %b %Y, %H:%M"))

    # Empty hash is passed to create the empty model with just model name the first time
    # Complete hash is passed to create the model with all the details in the second time
    if hash == "":
        position = len(config[config_keys.model.value]) + 1

        config[config_keys.model.value][position] = {
            "name": name,
            "desc": str(description),
            "time": current_datetime,
            "hash": hash,
            "version": version,
            "code": code,
        }
    else:
        position = len(config[config_keys.model.value])
        model_name_position = config[config_keys.model.value][position]["name"]
        if model_name_position == name:
            config[config_keys.model.value][position]["hash"] = hash
            config[config_keys.model.value][position]["version"] = version
            config[config_keys.model.value][position]["code"] = code

    save_config(config=config)


def add_metrics_to_config(
    values, model_name=None, model_version=None, func=None
):
    config = load_config()

    if model_name is None:
        model_name, model_version, model_hash = get_model_latest(
            config=config
        )

    if len(config[config_keys.metrics.value]) != 0:
        metric_values = config[config_keys.metrics.value]["values"]
        # metric_values = update_step_dict(metric_values, values)
        metric_values.update(values)
        # print('default',metric_values)
    else:
        metric_values = values

        # print('not default',metric_values)

    hash = generate_hash_for_dict(values=metric_values)

    current_datetime = datetime.now()
    current_datetime = str(current_datetime.strftime("%d %b %Y, %H:%M"))

    config[config_keys.metrics.value].update(
        {
            "values": metric_values,
            "hash": hash,
            "model_name": model_name,
            "model_version": model_version,
            "time": current_datetime,
        }
    )

    save_config(config=config)


def load_metrics_from_config():

    config = load_config()
    try:
        metrics = config[config_keys.metrics.value]["values"]
    except Exception as e:
        # print(e)
        print("No metrics are found in config")
        metrics = {}

    return metrics


def add_params_to_config(
    values, model_name=None, model_version=None, func=None
):
    config = load_config()

    if model_name is None:
        model_name, model_version, model_hash = get_model_latest(
            config=config
        )

    if len(config[config_keys.params.value]) != 0:
        param_values = config[config_keys.params.value]["values"]
        # param_values = update_step_dict(param_values, values)
        param_values.update(values)
    else:
        param_values = values

    hash = generate_hash_for_dict(values=param_values)
    # print("params", model_version)

    current_datetime = datetime.now()
    current_datetime = str(current_datetime.strftime("%d %b %Y, %H:%M"))

    config[config_keys.params.value].update(
        {
            "values": param_values,
            "hash": hash,
            "model_name": model_name,
            "model_version": model_version,
            "time": current_datetime,
        }
    )

    save_config(config=config)


def load_params_from_config():

    config = load_config()
    try:
        metrics = config[config_keys.params.value]["values"]
    except Exception as e:
        # print(e)
        print("No params are found in config")
        metrics = {}

    return metrics


def add_figures_to_config(
    values, model_name=None, model_version=None, func=None
):
    config = load_config()

    if model_name is None:
        model_name, model_version, model_hash = get_model_latest(
            config=config
        )

    if len(config[config_keys.figure.value]) != 0:
        figure_values = config[config_keys.figure.value]["values"]
        figure_values.update(values)
    else:
        figure_values = values

    hash = generate_hash_for_dict(values=figure_values)
    # print("figures", model_version)

    current_datetime = datetime.now()
    current_datetime = str(current_datetime.strftime("%d %b %Y, %H:%M"))

    config[config_keys.figure.value].update(
        {
            "values": figure_values,
            "hash": hash,
            "model_name": model_name,
            "model_version": model_version,
            "time": current_datetime,
        }
    )

    save_config(config=config)


def load_figures_from_config():

    config = load_config()
    try:
        figures = config[config_keys.figure.value]["values"]
    except Exception as e:
        # print(e)
        print("No figures are found in config")
        figures = {}

    return figures


def add_pred_to_config(
    values, model_name=None, model_version=None, func=None
):
    config = load_config()

    if model_name is None:
        model_name, model_version, model_hash = get_model_latest(
            config=config
        )

    if len(config[config_keys.pred_function.value]) != 0:
        # pred_function_values = config["pred_function"]["values"]
        # pred_function_values.update(values)
        pred_function_values = values
    else:
        pred_function_values = values

    hash = generate_hash_for_dict(values=pred_function_values)
    # print("pred_function", model_version)

    current_datetime = datetime.now()
    current_datetime = str(current_datetime.strftime("%d %b %Y, %H:%M"))

    config[config_keys.pred_function.value].update(
        {
            "values": pred_function_values,
            "hash": hash,
            "model_name": model_name,
            "model_version": model_version,
            "time": current_datetime,
        }
    )

    save_config(config=config)


def load_pred_from_config():

    config = load_config()
    try:
        pred_file = config[config_keys.pred_function.value]["values"]
    except Exception as e:
        # print(e)
        print("No pred_functions are found in config")
        pred_file = {}

    return pred_file


def add_pip_req_to_config(
    values, model_name=None, model_version=None, func=None
):
    config = load_config()

    if model_name is None:
        model_name, model_version, model_hash = get_model_latest(
            config=config
        )

    if len(config[config_keys.pip_requirement.value]) != 0:
        # pip_requirement_values = config["pip_requirement"]["values"]
        # pip_requirement_values.update(values)
        pip_requirement_values = values
    else:
        pip_requirement_values = values

    hash = generate_hash_for_dict(values=pip_requirement_values)
    # print("pred_function", model_version)

    current_datetime = datetime.now()
    current_datetime = str(current_datetime.strftime("%d %b %Y, %H:%M"))

    config[config_keys.pip_requirement.value].update(
        {
            "values": pip_requirement_values,
            "hash": hash,
            "model_name": model_name,
            "model_version": model_version,
            "time": current_datetime,
        }
    )

    save_config(config=config)


def load_pip_req_from_config():

    config = load_config()
    try:
        pip_req_file = config[config_keys.pip_requirement.value]["values"]
    except Exception as e:
        # print(e)
        print("No pip_requirement are found in config")
        pip_req_file = {}

    return pip_req_file


def add_resource_to_config(
    values, model_name=None, model_version=None, func=None
):
    config = load_config()

    if model_name is None:
        model_name, model_version, model_hash = get_model_latest(
            config=config
        )

    if len(config[config_keys.resource.value]) != 0:
        # pip_requirement_values = config["pip_requirement"]["values"]
        # pip_requirement_values.update(values)
        pip_requirement_values = values
    else:
        pip_requirement_values = values

    hash = generate_hash_for_dict(values=pip_requirement_values)
    # print("pred_function", model_version)

    current_datetime = datetime.now()
    current_datetime = str(current_datetime.strftime("%d %b %Y, %H:%M"))

    config[config_keys.resource.value].update(
        {
            "values": pip_requirement_values,
            "hash": hash,
            "model_name": model_name,
            "model_version": model_version,
            "time": current_datetime,
        }
    )

    save_config(config=config)


def load_resource_from_config():

    config = load_config()
    try:
        pip_req_file = config[config_keys.resource.value]["values"]
    except Exception as e:
        # print(e)
        print("No resource are found in config")
        pip_req_file = {}

    return pip_req_file


def add_artifacts_to_config(name, values, func):
    hash = ""
    version = ""
    config = load_config()

    model_name, model_version, model_hash = get_model_latest(
        config=config
    )

    current_datetime = datetime.now()
    current_datetime = str(current_datetime.strftime("%d %b %Y, %H:%M"))

    position = len(config[config_keys.artifacts.value]) + 1
    config[config_keys.artifacts.value][position] = {
        "name": name,
        "hash": hash,
        "version": version,
        "model_name": model_name,
        "model_version": model_version,
        "time": current_datetime,
    }


def get_model_latest(config, version="latest"):
    config_model = config[config_keys.model.value]
    model_name = None
    model_version = None
    model_hash = None

    model_positions = list(config_model.keys())

    if len(model_positions) != 0:
        # print(model_positions)
        position = model_positions[-1]
        model_name = config_model[position]["name"]
        model_version = config_model[position]["version"]
        model_hash = config_model[position]["hash"]

    return model_name, model_version, model_hash



from pydantic import BaseModel
from pureml.components import model
from rich import print


def get_model_helper(label_model):
        model_load = model.fetch(label = label_model)
        if model_load is None:
            print("[orange] Unable to fetch the model")
            return None
        else:
            print("[green] Succesfully fetched the model")
            return model_load



class StorageProvider:
    def upload_file(self, source_file, destination) -> str or None:
        """
        Upload a file to the provided relative destination path

        Returns - relative file path (str)
        """
        raise NotImplementedError("Subclasses must implement this method")

    def download_file(self, source, destination_file) -> None:
        """
        Download a file from the relative source path to the destination path

        Returns None
        """
        raise NotImplementedError("Subclasses must implement this method")

    def exists(self, file_path) -> bool:
        """
        Checks whether the file path exists

        Returns - True if exists (bool)
        """
        raise NotImplementedError("Subclasses must implement this method")


import typer
from pureml.cli.puremlconfig import PureMLConfigYML
import typer
from rich import print
from rich.table import Table

# from puretrainer.train import Trainer
import os
from dotenv import load_dotenv
from pathlib import Path
from pureml.cli.helpers import get_auth_headers
from pureml.utils.logger import get_logger

load_dotenv()


import pureml.cli.auth as auth
import pureml.cli.secrets as secrets
import pureml.cli.orgs as orgs

logger = get_logger(name="sdk.cli.main")
app = typer.Typer(no_args_is_help=True)
app.add_typer(auth.app, name="auth")
app.add_typer(secrets.app, name="secrets")
app.add_typer(orgs.app, name="orgs")


def print_version(value: bool):
    if value:
        from pureml import __version__

        print(f"PureML SDK version: {__version__}")
        raise typer.Exit()


@app.callback()
def callback(
    version: bool = typer.Option(
        None, "--version", "-v", callback=print_version, is_eager=True
    ),
):
    """
    PureML CLI

    This is the official CLI for PureML.
    """


# init the config file
@app.command()
def init(
    silent: bool = typer.Option(False, "--silent", "-s", help="Run in silent mode"),
    repository: str = typer.Option(
        None, "--repository", "-r", help="The repository to use"
    ),
    backend_url: str = typer.Option(
        None, "--backend-url", "-b", help="The backend url to use"
    ),
    frontend_url: str = typer.Option(
        None, "--frontend-url", "-f", help="The frontend url to use"
    ),
):
    """
    This command will initialize the puremlconfig.yaml file for your project
    """
    project_path = Path.cwd()
    puremlconfig = PureMLConfigYML(project_path / "puremlconfig.yaml")
    headers = get_auth_headers()
    if headers is None:
        return

    # check if the config file exists
    if puremlconfig.file.exists():
        logger.warning(
            "A puremlconfig.yaml file already exists in this directory - aborting"
        )
        print("A puremlconfig.yaml file already exists in this directory - aborting")
        return

    # if not silent, ask for the config values
    if not silent:
        print("Let's get started with the configuration of your project")

        repository_path = ""
        if not repository:
            table = Table("Sr.No.", "Storage Repository")
            table.add_row("1 (default)", "Local file system")
            table.add_row("2", "AWS S3")
            table.add_row("3", "Cloudflare R2")
            print(table)
            repository = None
            while repository not in [1, 2, 3]:
                repository = typer.prompt(
                    "Enter the Sr.No. of your preferred storage repository (1, 2 or 3)",
                    default=1,
                    show_default=False,
                )

            logger.info(f"Selected repository: {repository}")
            if repository == 1:
                path = typer.prompt(
                    "Enter the path to your local repository [default: pureml_data/]",
                    default="pureml_data/",
                    show_default=False,
                )
                repository_path = "file://" + path
            elif repository == 2 or repository == 3:
                print(
                    "If not already done, please integrate your storage repository by adding secrets using `pureml secrets add`"
                )
                added_secret = typer.confirm(
                    "Have you added your storage repository secrets? (y/n)",
                    default=True,
                )
                if not added_secret:
                    print(
                        "Please add your storage repository secrets using `pureml secrets add` or using UI"
                    )
                    return
                secret_name = typer.prompt("Enter the name of your integrated secret")
                repository_path = f"{repository}://{secret_name}"

        else:
            print(f"Using given repository: {repository}")

        isSelfHosted = typer.confirm("Are you using a self-hosted PureML instance?")
        logger.info(f"Is self-hosted: {isSelfHosted}")
        if isSelfHosted:
            if not backend_url:
                backend_url = typer.prompt(
                    "Enter the backend url", default="http://localhost:8080/api/v1/"
                )
            else:
                print(f"Using given backend url: {backend_url}")

            if not frontend_url:
                frontend_url = typer.prompt(
                    "Enter the frontend url", default="http://localhost:3000/"
                )
            else:
                print(f"Using given frontend url: {frontend_url}")
    else:
        if not repository:
            print("Missing required arguments")
            return
        print(f"Using repository: {repository_path}")
        if backend_url:
            print(f"Using given backend url: {backend_url}")
        if frontend_url:
            print(f"Using given frontend url: {frontend_url}")

    # create the config file
    data = {
        "repository": repository_path,
    }
    if backend_url:
        data["backend_url"] = backend_url
    if frontend_url:
        data["frontend_url"] = frontend_url
    puremlconfig.save(data)
    print(
        "[green]Created configuration file puremlconfig.yaml in current directory[/green]"
    )



from pureml.schema import StorageSchema


def set_storage(storage_key: str) -> None:
    """Set the storage key.

    Args:
        storage (str): The storage key to reference storage details added in secret.
    """
    storage: StorageSchema = StorageSchema().get_instance()
    storage.STORAGE = storage_key
    # print(storage)
    # print(StorageSchema().get_instance())



from pydantic import BaseModel
from typing import ClassVar

from .backend import BackendSchema
from .paths import PathSchema
import os


class DatasetSchema(BaseModel):

    #_paths: PathSchema = PathSchema().get_instance()
    #_backend: BackendSchema = BackendSchema().get_instance() 

    # When Executed the above lines, the following error is thrown:  Is it  pydantic.fields.ModelPrivateAttr
    #So, Created a Class Instance and Accessed them

    _paths : PathSchema = PathSchema()
    _backend : BackendSchema = BackendSchema()
    PATH_DATASET_README:ClassVar = os.path.join(_paths.PATH_DATASET_DIR, "ReadME.md")  
    # Since PathSchema is ModelPrivateAttr, it cannot be named directly, We need to use "_" infront of variable name. So, _paths is used instead of paths.

    #Required Type Annotation when accessing ModelPrivateAttr, so used ClassVar



import pandas as pd
import numpy as np
from PIL import Image
import json
from pureml.schema import DataTypes


def process_input(input):

    input_type = ""
    input_shape = None

    if input is not None:
        input_keys = input.keys()

        if "type" in input_keys:
            input_type = input["type"]
        if "shape" in input_keys:
            input_shape = input["shape"]

    return input_type, input_shape


def process_output(output):

    output_type = ""
    output_shape = None

    if output is not None:
        output_keys = output.keys()

        if "type" in output_keys:
            output_type = output["type"]
        if "shape" in output_keys:
            output_shape = output["shape"]

    return output_type, output_shape


def parse_input(data, input_type, input_shape):
    # if input_type == 'json':
    #     data = data

    if input_type == DataTypes.NUMPY_NDARRAY:
        if type(data) == str:
            data = json.loads(data)
        data = np.array(data)
        data = data.reshape(input_shape)
    elif input_type == DataTypes.PD_DATAFRAME:
        if type(data) == str:
            data = json.loads(data)
        data = pd.DataFrame.from_dict(data)
    elif input_type == DataTypes.TEXT:
        data = json.loads(data)
    elif input_type == DataTypes.IMAGE:
        data = Image.open(data)
        data = np.array(data)
        print("input image shape", data.shape)
    else:
        data = None

    print("input data type", type(data))
    # print(data)

    return data


def parse_output(data, output_type, output_shape):
    # if input_type == 'json':
    #     data = data

    if output_type == DataTypes.NUMPY_NDARRAY:
        data = data.tolist()
        data = json.dumps(data)
    elif output_type == DataTypes.PD_DATAFRAME:
        data = data.to_dict()
        data = json.dumps(data)
    elif output_type == DataTypes.TEXT:
        data = json.dumps(data)
    else:
        data = json.dumps(data)

    return data



from dotenv import load_dotenv
import os
from pureml.schema import Env
from rich import print


def generate_env_dict(env_path):
    load_dotenv(dotenv_path=env_path)

    org_id = os.getenv(Env.ORG_ID.value)
    access_token = os.getenv(Env.ACCESS_TOKEN.value)
    api_token = os.getenv(Env.API_TOKEN.value)

    env_dict = {}

    if org_id is None:
        print("[orange]", Env.ORG_ID.value, "is not set in .env")
    else:
        env_dict[Env.ORG_ID.value] = org_id

    if access_token is None:
        print("[orange]", Env.ACCESS_TOKEN.value, "is not set in .env")
    else:
        env_dict[Env.ACCESS_TOKEN.value] = access_token

    if api_token is None:
        print("[orange]", Env.API_TOKEN.value, "is not set in .env")
    else:
        env_dict[Env.API_TOKEN.value] = api_token

    return env_dict


def validate_env_docker(env_dict):
    env_dict_set = set(list(env_dict.keys()))
    env_dict_docker = set([Env.ORG_ID.value, Env.API_TOKEN.value])

    if env_dict_docker.issubset(env_dict_set):
        return True

    return False



import os


# BASE_URL = 'https://dev-api.pureml.com'
# BASE_URL = 'https://api.pureml.com'
# BASE_URL = 'https://pureml-development.up.railway.app/'
# BASE_URL = "https://pureml-development.up.railway.app/api/v1/"

# STORAGE = "PUREML-STORAGE"

# PATH_USER_TOKEN = os.path.join(os.path.expanduser("~"), ".pureml/token")

# PATH_PUREML_RELATIVE = ".pureml"

# PATH_USER_PROJECT_DIR = os.path.join(os.getcwd(), PATH_PUREML_RELATIVE)

# PATH_USER_PROJECT = os.path.join(PATH_USER_PROJECT_DIR, "pure.project")


# PATH_CONFIG = os.path.join(PATH_USER_PROJECT_DIR, "config.pkl")  # 'temp.yaml'

# PATH_ARTIFACT_DIR = os.path.join(PATH_USER_PROJECT_DIR, "artifacts")
# PATH_ARRAY_DIR = os.path.join(PATH_USER_PROJECT_DIR, "array")
# PATH_AUDIO_DIR = os.path.join(PATH_USER_PROJECT_DIR, "audio")
# PATH_FIGURE_DIR = os.path.join(PATH_USER_PROJECT_DIR, "figure")
# PATH_TABULAR_DIR = os.path.join(PATH_USER_PROJECT_DIR, "tabular")
# PATH_VIDEO_DIR = os.path.join(PATH_USER_PROJECT_DIR, "video")
# PATH_IMAGE_DIR = os.path.join(PATH_USER_PROJECT_DIR, "image")

# PATH_DATASET_DIR = os.path.join(PATH_USER_PROJECT_DIR, "dataset")
# PATH_DATASET_README = os.path.join(PATH_DATASET_DIR, "ReadME.md")


# PATH_MODEL_DIR = os.path.join(PATH_USER_PROJECT_DIR, "model")
# PATH_MODEL_README = os.path.join(PATH_MODEL_DIR, "ReadME.md")


# PATH_PREDICT_DIR_RELATIVE = "predict"
# PATH_PREDICT_DIR = os.path.join(PATH_USER_PROJECT_DIR, PATH_PREDICT_DIR_RELATIVE)
# PATH_PREDICT_REQUIREMENTS = os.path.join(PATH_PREDICT_DIR, "requirements.txt")
# PATH_PREDICT = os.path.join(PATH_PREDICT_DIR, "predict.py")

# PATH_PREDICT_USER = os.path.join(os.getcwd(), "predict.py")
# PATH_PREDICT_REQUIREMENTS_USER = os.path.join(os.getcwd(), "requirements.txt")


# PATH_FASTAPI_FILE = os.path.join(PATH_PREDICT_DIR, "fastapi_server.py")
# PORT_FASTAPI = 8005  # Same port as docker server


# PATH_DOCKER_DIR = os.path.join(PATH_USER_PROJECT_DIR,'docker')
# PATH_DOCKER_IMAGE = os.path.join(PATH_PREDICT_DIR, "Dockerfile")
# PATH_DOCKER_CONFIG = os.path.join(PATH_PREDICT_DIR, "DockerConfig.yaml")
# PORT_DOCKER = 8005      #Same port as fastapi server
# PORT_HOST = 8000
# BASE_IMAGE_DOCKER = "python:3.8-slim"
# API_IP_DOCKER = "0.0.0.0"
# API_IP_HOST = "0.0.0.0"



from pathlib import Path
from typing import Optional
import jwt
import requests

# import typer
from rich import print
from rich.syntax import Syntax

import os
import json
import typing

from urllib.parse import urljoin

from pureml.cli.helpers import get_auth_headers

from . import get_org_id

from pureml.schema import PathSchema, BackendSchema, ContentTypeHeader
from joblib import Parallel, delayed

#path_schema = PathSchema().get_instance()
#backend_schema = BackendSchema().get_instance()
path_schema = PathSchema()
backend_schema = BackendSchema()

def details(
    model_name: str, model_version: str = "latest", name: str = ""
):
    """This function returns the details of the artifact for a given model

    Parameters
    ----------
    model_name : str
        The name of the model you want to get the artifact details for
    model_version: str
        The version of the model
    name : str
        The name of the artifact.

    """

    org_id = get_org_id()

    url = "org/{}/model/{}/version/{}/log".format(
        org_id, model_name, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.status_code == 200:
        res_text = json.loads(response.text)

        if len(res_text) == 0:
            print("[yellow] No Artifacts have been found for the model")
            print(res_text)
            return
        else:
            print("[green]Artifacts have been found for the model")
            print(res_text)
            return res_text

    else:
        print("[orange]Unable to obtain the artifact details")
        print(response.text)
        return


# @app.command()
def add(
    artifact: str,
    name: str,
    model_name: str,
    model_version: str = "latest",
) -> str:
    """`add` function takes in the path of the artifact, name of the artifact and the model name and
    registers the artifact

    Parameters
    ----------
    artifact : str
        The path to the artifact file.
    name : str
        The name of the artifact.
    model_name : str
        The name of the model you want to add artifacts to.
    model_version: str
        The version of the model

    Returns
    -------
        The response is a JSON object

    """

    org_id = get_org_id()

    url = "org/{}/model/{}/version/{}/log".format(
        org_id, model_name, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    if os.path.isfile(artifact):
        file = {"file": (name, open(artifact, "rb"))}
    else:
        print("[orange] Artifact does not exist at the given path")

    data = {"name": name, "path": artifact}

    response = requests.post(url, data=data, files=file, headers=headers)

    if response.status_code == 200:
        print(f"[green]Artifacts have been registered!")

    else:
        print(f"[orange]Artifacts have not been registered!")
        print(response.text)

    return response.text


# @app.command()
def fetch(model_name: str, model_version: str = "latest", name: str = ""):
    """It fetches the artifact from the server and stores it in the local directory

    Parameters
    ----------
    model_name : str
        The name of the model you want to fetch the artifact from.
    model_version: str
        The version of the model
    name : str
        The name of the artifact to be fetched. If not specified, all artifacts will be fetched.

    Returns
    -------
        The response text is being returned.

    """

    org_id = get_org_id()

    def fetch_artifact(artifact_details: dict):

        url = artifact_details["location"]
        file_path_temp = artifact_details["path"]
        file_name = file_path_temp.split(os.path.sep)[-1]
        save_path = os.path.join(path_schema.PATH_ARTIFACT_DIR, file_name)
        print("save path", save_path)

        name_fetched = artifact_details["artifact"]

        headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

        print("Artifact url", url)

        # response = requests.get(url, headers=headers)
        response = requests.get(url)

        print(response.status_code)

        if response.status_code == 200:
            print("[green] Artifact {} has been fetched".format(name_fetched))

            save_dir = os.path.dirname(save_path)

            os.makedirs(save_dir, exist_ok=True)

            artifact_bytes = response.content

            open(save_path, "wb").write(artifact_bytes)

            print(
                "[green] Artifact {} has been stored at {}".format(
                    name_fetched, save_path
                )
            )

            return response.text
        else:
            print("[orange] Unable to fetch the artifact")

            return response.text

    artifact_details = details(
        model_name=model_name, name=name, model_version=model_version
    )

    if artifact_details is None:
        return

    if type(artifact_details) is dict:

        res_text = fetch_artifact(artifact_details)

    elif type(artifact_details) is list:
        res_text = Parallel(n_jobs=-1)(
            delayed(fetch_artifact)(art_det) for art_det in artifact_details
        )

    return res_text


# @app.command()
def delete(
    name: str, model_name: str, model_version: str = "latest"
) -> str:
    """`delete()` deletes an artifact from a model

    Parameters
    ----------
    name : str
        The name of the artifact you want to delete.
    model_name : str
        The name of the model you want to delete the artifact from
    model_version: str
        The version of the model

    """

    org_id = get_org_id()

    url = "org/{}/model/{}/version/{}/log".format(
        org_id, model_name, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    # artifact_details = details(model_name=model_name, artifact=artifact)

    # if artifact_details is None:
    #     print('[orange] Unable to find artifact details')
    #     return

    response = requests.delete(url, headers=headers)

    if response.status_code == 200:
        print(f"[green]Artifact has been deleted")

    else:
        print(f"[orange]Unable to delete artifact")

    return response.text


# if __name__ == "__main__":
#     app()



import os
import shutil

from ..provider import StorageProvider

class LocalStorageProvider(StorageProvider):
    def __init__(self, base_path: str):
        mod_base_path = os.path.join(base_path, "pureml_storage")
        if not os.path.exists(mod_base_path):
            os.makedirs(mod_base_path)
        self.base_path = mod_base_path

    def upload_file(self, source_file, relative_destination):
        destination = os.path.join(self.base_path, relative_destination)
        shutil.copy(source_file, destination)

    def download_file(self, relative_source, destination_file):
        source = os.path.join(self.base_path, relative_source)
        shutil.copy(source, destination_file)

    def exists(self, file_path) -> bool:
        return os.path.exists(file_path)


from pydantic import ConfigDict, BaseModel, root_validator
from .singleton import Singleton_BaseModel
import typing
import os
import typer

from pathlib import Path
from pureml.cli.puremlconfig import PureMLConfigYML


project_path = Path.cwd()
if Path.exists(project_path / "puremlconfig.yaml"):
    puremlconfig = PureMLConfigYML(project_path / "puremlconfig.yaml")
else:
    puremlconfig = None


def get_backend_base_url(backend_url: str = None):
    if (
        backend_url is not None
        and backend_url != ""
        and type(backend_url) == typer.Option
    ):
        # override backend url (command specific option)
        return backend_url
    if puremlconfig is not None:
        # user configured backend url (self-hosted or custom pureml backend instance)
        backend_url = (
            puremlconfig.data["backend_url"]
            if "backend_url" in puremlconfig.data
            else "https://pureml-development.up.railway.app/api/v1/"
        )
    else:
        # default backend url (production cloud backend)
        backend_url = "https://pureml-development.up.railway.app/api/v1/"
    return backend_url


def get_frontend_base_url(frontend_url: str = None):
    if (
        frontend_url is not None
        and frontend_url != ""
        and type(frontend_url) == typer.Option
    ):
        # override frontend url (command specific option)
        return frontend_url
    if puremlconfig is not None:
        # user configured frontend url (self-hosted or custom pureml frontend instance)
        frontend_url = (
            puremlconfig.data["frontend_url"]
            if "frontend_url" in puremlconfig.data
            #else "https://pureml.com/"
            else "https://app.pureml.com/"
        )
    else:
        # default frontend url (production cloud frontend)
        frontend_url =  "https://app.pureml.com/"
        #frontend_url = "https://pureml.com/"
    return frontend_url


class BackendSchema(Singleton_BaseModel):

    BASE_URL: str = get_backend_base_url()
    FRONTEND_BASE_URL: str = get_frontend_base_url()
    INTEGRATIONS: dict = {
        "s3": {
            "name": "AWS S3 Object Storage",
            "secrets": [
                "access_key_id",
                "access_key_secret",
                "bucket_location",
                "bucket_name",
            ],
        },
        "r2": {
            "name": "Cloudflare R2 Object Storage",
            "secrets": [
                "access_key_id",
                "access_key_secret",
                "account_id",
                "bucket_name",
                "public_url",
            ],
        },
    }
    model_config = ConfigDict(arbitrary_types_allowed=True)



from pureml.components import dataset
from rich import print



def get_dataset_helper(label_dataset):
    dataset_label = dataset.fetch(label_dataset)

    if dataset_label is None:
        print("[orange] Unable to fetch the dataset")
        return None
    else:
        print("[green] Succesfully fetched the dataset")
        return dataset_label


from urllib.parse import urljoin
import requests
import json
from rich import print
from pureml.cli.auth import save_auth
from . import delete_token, get_api_token, get_org_id
from pureml.schema import BackendSchema
from pureml.utils.logger import get_logger

# backend_schema = BackendSchema().get_instance()
backend_schema = BackendSchema()
logger = get_logger("sdk.components.auth")


def login(org_id: str, api_token: str) -> str:
    """The function takes in a user API token and logs in a user for a session.

    Parameters
    ----------
    api_token: str
        API Key of the token to be used for login

    """
    logger.info("Logging in user")
    logger.debug("Logging in user", org_id=org_id, api_token=api_token)

    url_path_1 = "org/id/{}".format(org_id)
    url = urljoin(backend_schema.BASE_URL, url_path_1)

    if api_token is None or org_id is None or api_token == "" or org_id == "":
        print("[orange]Invalid credentials for login")
        return

    headers = {"X-Api-Key": api_token}

    response = requests.get(url, headers=headers)
    logger.debug("Response from login", response_status=response.status_code)

    # print(response.text)
    if response.status_code == 200:

        response_text = response.text
        response_org_details = json.loads(response_text)["data"]

        # if response_org_details is not None:
        response_org_id = response_org_details[0]["uuid"]

        if response_org_id == org_id:
            print("[green]Valid Org Id and API token. Logged in successfully")
            save_auth(org_id=org_id, api_token=api_token)

        else:
            print("[orange]Valid Org Id and API token. Obtained different organization")

        # else:
        #     print('[green] Invalid Org Id and Access token')
    elif response.status_code == 403:
        print("[orange]Invalid API token")
        # delete_token(True)
    elif response.status_code == 404:
        print("[orange]Invalid Org Id")
        # delete_token(True)
    else:
        print("[orange]Unable to obtain the organization details")
        # delete_token(True)



def parse_version_label(label: str):
    name = None
    version = "latest"

    if label is None:
        return name, version

    separator_count = label.count(":")

    if separator_count == 0:
        name = label
    elif separator_count == 1:
        name, version = label.split(":")
    else:
        raise Exception("Invalid label format")

    return name, version


def validate_name(name: str):
    pass

def validate_version(version: str):
    pass


def generate_label(name: str, version: str = "latest"):
    label = ":".join([name, version])

    return label



from pathlib import Path
from urllib.parse import urljoin
from uuid import uuid4
from shutil import copyfileobj

import requests
from pureml.cli.helpers import get_auth_headers
from pureml.components import get_org_id

from pureml.schema.backend import BackendSchema
from pureml.schema.request import AcceptHeader
from .providers.aws import S3StorageProvider
from .providers.cloudflare import R2StorageProvider
from .providers.local import LocalStorageProvider

from pureml.cli.puremlconfig import PureMLConfigYML


project_path = Path.cwd()
if Path.exists(project_path / "puremlconfig.yaml"):
    puremlconfig = PureMLConfigYML(project_path / "puremlconfig.yaml")
else:
    puremlconfig = None

backend_schema = BackendSchema().get_instance()


def upload_and_get_provider_and_path(file_path: str, storage_provider: str = None, opt_base_dir: str = "") -> str or None:
    """
    Gets the storage provider from config repository and uploads the file to the storage provider

    Config file should have the following format -
    repository: <storage_provider>://<repository_argument>

    repository_argument is the argument required for the storage provider
    for file : repository_argument is the path of the directory where the file is to be stored
    for s3 : repository_argument is the name of the secret made in pureml secrets for s3
    for r2 : repository_argument is the name of the secret made in pureml secrets for r2


    Parameters
    ----------
    file_path: str
        Path of the file to be uploaded

    storage_provider: str (default: None)
        Storage provider to be used for uploading the file

    Returns
    -------
    str, str or None
        Provider used and Path of the uploaded file or None if the file is not uploaded successfully
    """
    repository_argument = ""
    if storage_provider is None:
        if puremlconfig is not None:
            storage_provider = puremlconfig.data["repository"].split("://")[0]
            repository_argument = puremlconfig.data["repository"].split(
                "://")[1]
        else:
            raise KeyError("No storage provider found in config file. Please run `pureml init` to initialize the config file or update the `puremlconfig.yaml` config file manually")

    if storage_provider not in ["s3", "r2", "file"]:
        raise ValueError("Invalid storage provider. Currently supported storage providers are s3, r2 and file")

    if storage_provider == "file":
        # if file storage is used, copy the file from given path to the repository_argument path
        # and return the path of the file
        file_path = Path(file_path)
        if not file_path.exists():
            raise FileNotFoundError("File not found")
        if not Path(repository_argument).exists():
            # create the directory if it doesn't exist
            Path(repository_argument).mkdir(parents=True, exist_ok=True)
        file_path = file_path.resolve()

        # copy the file to the repository_argument path
        copyfileobj(file_path.open("rb"), (Path(repository_argument) / file_path.name).open("wb"))
        
        # return the path of the file
        file_path = Path(repository_argument) / file_path.name
        return storage_provider, file_path.as_posix()

    # windows paths support and appending base dir
    file_path = file_path.replace(r'\\', "/").lstrip("/")

    # Get secrets from backend API
    org_id = get_org_id()
    url = f"org/{org_id}/secret/{str(repository_argument)}"
    url = urljoin(backend_schema.BASE_URL, url)
    # print(url)

    headers = get_auth_headers(content_type=None, accept=AcceptHeader.APP_JSON)

    response = requests.get(url, headers=headers)

    if response.ok:
        secrets = response.json()["data"]
        if secrets and len(secrets) != 0:
            secrets = secrets[0]
    else:
        # print(response.json())
        raise ValueError("Invalid repository argument")

    # Get storage provider client
    del secrets["source_type"]
    storage_provider_client = get_storage_provider(storage_provider, **secrets)

    key = file_path.split("/")[-1]
    ext = key.split(".")[-1]
    name = ".".join(key.split(".")[:-1])
    random_string = str(uuid4())[:8]
    key = f"{opt_base_dir}/{name}_{random_string}.{ext}"

    upload_path = storage_provider_client.upload_file(file_path, key)
    print("upload_path", upload_path)

    if upload_path is None:
        raise ValueError("File not uploaded successfully")

    return storage_provider, upload_path


def get_storage_provider(provider, **kwargs):
    """
    Function to dynamically create a client for a storage provider
    NOTE: all keyword arguments are passed from backend secrets hence are defined in the providers although not used
    Currently supported storage providers -

    ---

    1. S3
    ...

    Required kwargs -

    access_key_id: str
        AWS access key

    access_key_secret: str
        AWS secret key

    bucket_name: str
        AWS bucket name

    bucket_location: str
        AWS region

    public_url: str (default: None)
        Public URL of the bucket

    2. R2
    ...

    Required kwargs -

    access_key_id: str
        Cloudflare access key

    access_key_secret: str
        Cloudflare secret key

    bucket_name: str
        Cloudflare bucket name

    account_id: str
        Cloudflare account ID

    public_url: str
        Public URL of the bucket
    """
    if provider == "s3":
        return S3StorageProvider(**kwargs)
    elif provider == "r2":
        return R2StorageProvider(**kwargs)
    # elif provider == "local":
    #     return LocalStorageProvider(**kwargs)
    else:
        raise ValueError("Invalid storage provider")



from pydantic import BaseModel
from .paths import PathSchema
from .backend import BackendSchema
import os
from typing import ClassVar

class ModelSchema(BaseModel):
    #_paths: PathSchema = PathSchema().get_instance()
    #_backend: BackendSchema = BackendSchema().get_instance()
    _paths: PathSchema = PathSchema()
    _backend: BackendSchema = BackendSchema()
    PATH_MODEL_README: ClassVar = os.path.join(_paths.PATH_MODEL_DIR, "ReadME.md")
    PATH_MODEL: ClassVar = os.path.join(_paths.PATH_MODEL_DIR, "ran.pkl")



from pydantic import BaseModel
from pureml.predictor.predictor import BasePredictor
from importlib import import_module
from rich import print



class get_predictor_helper(BaseModel):
    predictor: BasePredictor = None
    predictor_path: str = "predict.py"

    def load_predictor(self):
        module_path = self.predictor_path.replace(".py", "")
        module_import = import_module(module_path)

        predictor_class = getattr(module_import, "Predictor")

        self.predictor = predictor_class()
        print(f"self.predictor: {self.predictor}")
        print("[green] Succesfully fetched the predictor")

        return self.predictor if self.predictor != None else None



from .get_dataset import get_dataset_helper
from .get_model import get_model_helper
from .get_predictor import get_predictor_helper
import requests
from .routes import *
from pydantic import ConfigDict, BaseModel
from rich import print
from typing import Any
from pureml_evaluate.evaluators.evaluator import eval as eval_fn
from pureml.cli.auth import get_auth_headers
from pureml.components import get_org_id
from pureml.predictor.predictor import BasePredictor
import json
import httpx


class EvalHelper(BaseModel):
    label_model: str
    label_dataset: str
    dataset: Any = None
    task_type: None = None
    model: Any = None
    predictor: Any = None
    model_config = ConfigDict(validate_assignment=True, arbitrary_types_allowed=True)

    def load_dataset(self):
        self.dataset = get_dataset_helper(self.label_dataset)

    def load_model(self):
        self.predictor.load_models()
        # self.model = get_model_helper(self.label_model)
        # print(f"self.model : {self.predictor.load_models()}")

    def load_predictor(self):
        predictor_class = get_predictor_helper()
        self.predictor = predictor_class.load_predictor()
        # print(f"From Evaluator File: {self.predictor}")

    def load(self):
        self.load_dataset()
        self.load_predictor()
        self.load_model()

    def get_y_pred(self):
        return self.predictor.predict(self.dataset["x_test"]) if ('x_test' in self.dataset.keys() and self.predictor != None) else None

    def get_y_true(self):
        return self.dataset["y_test"] if 'y_test' in self.dataset.keys() else None

    def get_sensitive_features(self):
        return self.dataset['sensitive_features'] if 'sensitive_features' in self.dataset.keys() else None


def evalfunction(label_model: str, label_dataset: str, task_type: list[str], metrics=None, path_to_config=None, pdf_file_name='metrics_graph.pdf'):
    evaluator = EvalHelper(
        task_type=None, label_dataset=label_dataset, label_model=label_model
    )

    evaluator.load()

    y_pred = evaluator.get_y_pred()
    y_true = evaluator.get_y_true()
    sensitive_features = evaluator.get_sensitive_features()

    values = eval_fn(y_true=y_true, y_pred=y_pred, sensitive_features=sensitive_features,
                     evaluators=task_type, y_pred_scores=None, metrics=metrics,
                     path_to_config=path_to_config, as_pdf=True, pdf_file_name=pdf_file_name)

    # send_data_to_backend(values, label_model)
    return values


def send_data_to_backend(values, label_model):
    keys_in_values = values.keys()
    orgID = get_org_id()
    model_name, branch_name, version = label_model.split(":")
    risk_url = ModelRisk.format(
        orgId=orgID, modelName=model_name, branchName=branch_name, version=version)
    print(f"risk_url : {risk_url}")
    for i in keys_in_values:            # i = complete / subsets
        categories = values[i].keys()
        for j in categories:             # j = performance / fairness
            metrics = values[i][j].keys()
            for k in metrics:              # k  = accuracy / precision
                payload = create_payload(values, i, j, k)
                print(payload)
                try:
                    response = httpx.post(
                        risk_url, data=payload, headers=get_auth_headers())
                    # response = requests.post(risk_url, data=payload,headers = get_auth_headers())
                    print(f"Status Code: {response.status_code}")
                except Exception as e:
                    print(e)

                if response.status_code == 200:
                    print(
                        "[green] Succesfully sent the evaluation results to PureML server")
                else:
                    print(response.status_code)
                    print(response.json())


def create_payload(values, i, j, k):
    try:
        value = format(values[i][j][k]['value'], '.2f')
        payload = {
            "category": f"{j}",
            "risk": f"{k}",
                    "severity": f"{values[i][j][k]['severity']}",
                    "value": str(value),
                    "threshold": str(values[i][j][k]['threshold']),
                    "policy": "test-sr-11-7"
        }
        return json.dumps(payload, indent=4)
    except Exception as e:
        print(e)
        return None



from pydantic import ConfigDict, BaseModel
from .backend import BackendSchema

import typing
import os
from .paths import PathSchema
from typing import ClassVar

class FastAPISchema(BaseModel):
    #paths: PathSchema = PathSchema().get_instance()
    paths: PathSchema = PathSchema()
    PATH_FASTAPI_FILE:ClassVar = os.path.join(paths.PATH_PREDICT_DIR, "fastapi_server.py")
    PORT_FASTAPI:ClassVar = 8005  # Same port as docker server
    API_IP_HOST:ClassVar = "0.0.0.0"
    model_config = ConfigDict(arbitrary_types_allowed=True)


class DockerSchema(BaseModel):
    #paths: PathSchema = PathSchema().get_instance()
    #backend: BackendSchema = BackendSchema().get_instance()
    paths: PathSchema = PathSchema()
    backend: BackendSchema = BackendSchema()
    PATH_DOCKER_IMAGE:ClassVar = os.path.join(paths.PATH_PREDICT_DIR, "Dockerfile")
    PATH_DOCKER_CONFIG:ClassVar = os.path.join(paths.PATH_PREDICT_DIR, "DockerConfig.yaml")
    PORT_DOCKER:ClassVar = FastAPISchema().PORT_FASTAPI  # Same port as fastapi server
    PORT_HOST:ClassVar = 8000
    BASE_IMAGE_DOCKER:ClassVar = "python:3.8-slim"
    API_IP_DOCKER:ClassVar = FastAPISchema().API_IP_HOST
    API_IP_HOST:ClassVar = "0.0.0.0"
    model_config = ConfigDict(arbitrary_types_allowed=True)



import typing
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path

import joblib
from pydantic import BaseModel


class ModelFrameworkType(Enum):
    SKLEARN = 'sklearn'
    XGBOOST = 'xgboost'
    LIGHTGBM = 'lightgbm'
    CATBOOST = 'catboost'
    PYTORCH_TABNET = 'pytorch_tabnet' #pytorch
    PYTORCH = 'torch' #pytorch
    KERAS = 'keras'
    TENSORFLOW = 'tensorflow'
    # HUGGINGFACE_TRANSFORMER = 'huggingface_transformer'
    CUSTOM = 'custom'








class ModelFramework(ABC):
    @abstractmethod
    def typ(self) -> ModelFrameworkType:
        pass


    # @abstractmethod
    # def model_metadata(self, model) -> typing.Dict[str, str]:
    #     pass

    def model_type(self, model) -> str:
        return 'Model'

    # @abstractmethod
    # def model_name(self, model) -> str:
    #     return None
    
    def get_framework_version(self) -> str:
        pass

    @abstractmethod
    def get_requirements(self, framework_name: str) -> str:
        pass

    def supports_model_class(self, model_class) -> bool:
        pass

    # @abstractmethod
    def predict(self):
        pass




# class ModelConfig(BaseModel):
#     model: typing.Any = None
#     model_name: str = None
#     model_framework: typing.Any = None
#     model_requirements: list = None



#     @staticmethod
#     def from_dict(model_config_dict: dict):
#         config = ModelConfig(   
#             model = model_config_dict['model'],
#             # model_name = model_config_dict['model_name'],
#             model_framework = model_config_dict['model_framework'],
#             model_requirements = model_config_dict['model_requirements'])

#         return config



#     def to_dict(self):
#         return {
#             'model': self.model,
#             # 'model_name': self.model_name,
#             'model_framework': self.model_framework,
#             'model_requirements': self.model_requirements
#         }

#     @staticmethod
#     def load_from_disk(model_config_path: Path):
#         return  ModelConfig.from_dict(joblib.load(model_config_path))


#     def save_to_disk(self, model_config_path:Path):
#         joblib.dump(self.to_dict(), model_config_path)

#         return model_config_path



import json
import os

import typer
from pureml.components import get_token, get_api_token
from pureml.schema import PathSchema, ContentTypeHeader, AcceptHeader
from pureml.utils.logger import get_logger
from rich import print

# path_schema = PathSchema().get_instance()
path_schema = PathSchema()
logger = get_logger(name="sdk.cli.helpers")


def save_auth(
    org_id: str = None,
    access_token: str = None,
    email: str = None,
    api_token: str = None,
):
    logger.info("Saving authentication token")
    token_path = path_schema.PATH_USER_TOKEN

    token_dir = os.path.dirname(token_path)
    os.makedirs(token_dir, exist_ok=True)

    # Read existing token
    if os.path.exists(token_path):
        logger.info("Updating exsisting token file", token_path=token_path)
        with open(token_path, "r") as token_file:
            token = json.load(token_file)

        if org_id is not None:
            token["org_id"] = org_id
        if access_token is not None:
            token["accessToken"] = access_token
        if api_token is not None:
            token["api_token"] = api_token
        if email is not None:
            if "email" in token and token["email"] != email:
                token["org_id"] = ""
            token["email"] = email
    else:
        token = {
            "org_id": org_id,
            "accessToken": access_token,
            "email": email,
            "api_token": api_token,
        }
        if org_id is None:
            token["org_id"] = ""

    token = json.dumps(token)

    with open(token_path, "w") as token_file:
        token_file.write(token)

    logger.info("Token saved", token_path=token_path)


# content_type: str = "application/x-www-form-urlencoded",
# accept: str = "application/json",
def get_auth_headers(
    content_type: str = ContentTypeHeader.APP_FORM_URL_ENCODED,
    accept: str = None,
):
    token = get_token()
    api_token = get_api_token()

    if token is None and api_token is None:
        print(
            f"[red]Authentication token and API token does not exist! Use `pureml auth login` to login."
        )
        typer.Exit()
        return None
    else:
        headers = {}
        if content_type is not None:
            headers["Content-Type"] = content_type.value
        if accept is not None:
            headers["Accept"] = accept.value

        if token is not None:
            headers["Authorization"] = "Bearer {}".format(token)
            return headers

        if api_token is not None:
            headers["X-Api-Key"] = api_token["api_token"]

            return headers



import typing
import numpy as np
import matplotlib.pyplot as plt
from . import metrics as pure_metrics
from . import params as pure_params
from . import figure as pure_figure
from . import predict as pure_predict
from . import pip_requirement as pure_pip_req
from pureml.utils.version_utils import parse_version_label
from pureml.schema import PathSchema, BackendSchema, LogSchema
import requests
from urllib.parse import urljoin
from . import get_org_id

#path_schema = PathSchema().get_instance()
#backend_schema = BackendSchema().get_instance()
path_schema = PathSchema()
backend_schema = BackendSchema()

post_keys = LogSchema().key


def log(label: str = None, metrics=None, params=None, step=1, **kwargs):

    if metrics is not None:
        func_params = {}

        if label is not None:
            func_params["label"] = label

        func_params["metrics"] = metrics.copy()

        func_params["step"] = step

        pure_metrics.add(**func_params)

    if params is not None:
        func_params = {}

        if label is not None:
            func_params["label"] = label

        func_params["params"] = params.copy()

        func_params["step"] = step

        pure_params.add(**func_params)

    if post_keys.figure.value in kwargs.keys():
        figure = kwargs[post_keys.figure.value]
        func_params = {}

        if label is not None:
            func_params["label"] = label

        func_params["figure"] = figure.copy()
        # func_params['step']  = step

        pure_figure.add(**func_params)

    if post_keys.predict.value in kwargs.keys():
        predict = kwargs[post_keys.predict.value]
        func_params = {}

        if label is not None:
            func_params["label"] = label

        func_params["path"] = predict

        pure_predict.add(**func_params)

    if post_keys.requirements.value in kwargs.keys():
        requirement = kwargs[post_keys.requirements.value]
        func_params = {}

        if label is not None:
            func_params["label"] = label

        func_params["path"] = requirement

        pure_pip_req.add(**func_params)



import json
import os
from urllib.parse import urljoin
import time
import joblib
import pandas as pd
import requests
from pureml.cli.helpers import get_auth_headers

from pureml.schema import (
    DatasetSchema,
    StorageSchema,
    ConfigKeys,
    ContentTypeHeader,
    AcceptHeader,
)
from pureml.storage import upload_and_get_provider_and_path
from pureml.utils.hash import generate_hash_for_file
from pureml.utils.readme import load_readme
from rich import print

from . import get_org_id
from pureml.utils.version_utils import parse_version_label
from pureml.utils.config import reset_config
from pureml.utils.logger import get_logger

config_keys = ConfigKeys
storage = StorageSchema().get_instance()
logger = get_logger("sdk.components.dataset")


def check_dataset_hash(hash: str, label: str):
    logger.info("Checking dataset hash", hash=hash, label=label)

    name, _ = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    url = "org/{}/dataset/{}/hash-status".format(org_id, name)
    url = urljoin(dataset_schema._backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    data = {"hash": hash}

    data = json.dumps(data)

    response = requests.post(url, data=data, headers=headers)

    hash_exists = False

    if response.ok:
        hash_exists = response.json()["data"][0]

    logger.info(f"Hash exists {hash_exists}")
    return hash_exists


def list():
    """This function will return a list of all the datasets

    Returns
    -------
        A list of all the datasets

    """
    logger.info("Fetching list of datasets")

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    url = "org/{}/dataset/all".format(org_id)
    url = urljoin(dataset_schema._backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        # print(f"[green]Obtained list of models[/green]")

        response_text = response.json()
        dataset_list = response_text["data"]
        # print(model_list)

        return dataset_list
    else:
        logger.error("Unable to obtain the list of dataset", response=response)
        print(f"[orange]Unable to obtain the list of dataset![/orange]")

    return


def init(label: str, readme: str = None):

    name, _ = parse_version_label(label)

    headers = get_auth_headers(content_type=ContentTypeHeader.ALL)
    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    if readme is None:
        readme = dataset_schema.PATH_DATASET_README

    file_content, file_type = load_readme(path=readme)

    url = "org/{}/dataset/{}/create".format(org_id, name)
    url = urljoin(dataset_schema._backend.BASE_URL, url)

    data = {
        "name": name,
        "readme": {"file_type": file_type, "content": file_content},
    }

    data = json.dumps(data)
    # files = {"file": (readme, open(readme, "rb"), file_type)}
    # response = requests.post(url, data=data, headers=headers, files=files)
    #
    response = requests.post(url, data=data, headers=headers)

    if response.ok:
        print(f"[green]Dataset has been created![/green]")

        return True

    else:
        print(response.text)
        print(f"[orange]Dataset has not been created![/orange]")

        return False


def save_dataset(dataset, name: str):
    logger.info("Saving dataset", name=name)
    dataset_schema = DatasetSchema()
    # file_name = '.'.join([name, 'parquet'])
    file_name = ".".join([name, "pkl"])
    save_path = os.path.join(dataset_schema._paths.PATH_DATASET_DIR, file_name)

    os.makedirs(dataset_schema._paths.PATH_DATASET_DIR, exist_ok=True)

    # dataset.to_parquet(save_path)
    joblib.dump(dataset, save_path)

    logger.info("Dataset saved", save_path=save_path)
    return save_path


def register(dataset, label: str, lineage, is_empty: bool = False) -> str:
    """The function takes in a dataset, a name and a version and saves the dataset locally, then uploads the
    dataset to the PureML server

    Parameters
    ----------
    dataset
        The dataset you want to register
    name : str
        The name of the dataset.
    version: str, optional
        The version of the dataset.

    """
    logger.info("Registering dataset", label=label, is_empty=is_empty)
    name, _ = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    dataset_path = save_dataset(dataset, name)

    # windows paths support
    if "\\" in dataset_path:
        dataset_path = dataset_path.replace("\\", "/")

    name_with_ext = dataset_path.split("/")[-1]

    dataset_hash = generate_hash_for_file(
        file_path=dataset_path, name=name, is_empty=is_empty
    )

    if is_empty:
        dataset_path = save_dataset(dataset=None, name=name)
        name_with_ext = dataset_path.split("/")[-1]

    dataset_exists = dataset_status(label)
    # print('Dataset status', dataset_exists)

    if not dataset_exists:
        dataset_created = init(label=label)
        # print('dataset_created', dataset_created)
        if not dataset_created:
            print("[orange] Unable to register the dataset[/orange]")
            return False, dataset_hash, None
    else:
        logger.info("Dataset already exists", label=label)
        print("[green] Connected to Dataset[/green]")

    dataset_exists_remote = check_dataset_hash(hash=dataset_hash, label=label)

    # print("Hash : ", dataset_hash)

    if dataset_exists_remote:

        print(f"[green]Dataset already exists. Not registering a new version![/green]")
        return True, dataset_hash, "latest"
    else:

        url = "org/{}/dataset/{}/register".format(org_id, name)
        url = urljoin(dataset_schema._backend.BASE_URL, url)

        headers = get_auth_headers(content_type=None, accept=AcceptHeader.APP_JSON)

        # files = {"file": (name_with_ext, open(dataset_path, "rb"))}
        provider, file_path = upload_and_get_provider_and_path(
            dataset_path, opt_base_dir="dataset-registry"
        )

        lineage = json.dumps(lineage)

        data = {
            "name": name,
            "hash": dataset_hash,
            "lineage": lineage,
            "is_empty": is_empty,
            "storage": provider,
            "file_path": file_path,
        }

        # data = json.dumps(data)

        response = requests.post(url, files={"foo": "bar"}, data=data, headers=headers)

        if response.ok:

            dataset_version = None

            # print(response.json())
            try:
                dataset_version = response.json()["data"][0]["version"]
                print("Dataset version: ", dataset_version)
                dataset_label = ":".join([name, dataset_version])
                print("Dataset label: ", dataset_label)

                if is_empty:
                    print(f"[green]Lineage has been registered![/green]")
                else:
                    print(f"[green]Dataset and lineage have been registered![/green]")

                # reset_config(key=config_keys.dataset.value)

            except Exception as e:
                logger.error(
                    "Incorrect json response. Dataset has not been registered", error=e
                )
                print(
                    "[orange]Incorrect json response. Dataset has not been registered[/orange]"
                )
                print(e)

            return True, dataset_hash, dataset_version
        else:
            logger.error("Unable to register dataset", response=response)
            print(f"[orange]Dataset has not been registered![/orange]")
            print(response.text)

            return True, dataset_hash, None


def dataset_status(label: str):

    name, _ = parse_version_label(label)

    dataset_details = details(label=label)

    if dataset_details:
        return True
    else:
        return False


def details(
    label: str,
):
    """It fetches the details of a dataset.

    Parameters
    ----------
    name : str
        The name of the dataset
    version: str
        The version of the dataset
    Returns
    -------
        The details of the dataset.

    """

    name, _ = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    url = "org/{}/dataset/{}".format(org_id, name)
    url = urljoin(dataset_schema._backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        # print(f"[green]Dataset details have been fetched[/green]")
        response_text = response.json()
        dataset_details = response_text["data"][0]

        return dataset_details

    else:
        print(f"[yellow]Dataset details have not been found[/yellow]")
        return


def version_details(label: str):
    """It fetches the details of a dataset.

    Parameters
    ----------
    name : str
        The name of the dataset
    version: str
        The version of the dataset
    Returns
    -------
        The details of the dataset.

    """

    name, version = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    url = "org/{}/dataset/{}/version/{}".format(org_id, name, version)
    url = urljoin(dataset_schema._backend.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        # print(f"[green]Dataset Version details have been fetched[/green]")
        response_text = response.json()
        dataset_details = response_text["data"][0]
        # print(dataset_details)

        return dataset_details

    else:
        print(f"[yellow]Dataset version details have not been found[/yellow]")
        return


def fetch(label: str):
    """This function fetches a dataset from the server and returns it as a dataframe object

    Parameters
    ----------
    name : str, optional
        The name of the dataset you want to fetch.
    version: str
        The version of the dataset

    Returns
    -------
        The dataset dataframe is being returned.

    """

    name, version = parse_version_label(label)

    org_id = get_org_id()
    dataset_schema = DatasetSchema()

    dataset_details = version_details(label=label)

    if dataset_details is None:
        print(f"[orange]Unable to fetch Dataset version[/orange]")
        return

    is_empty = dataset_details["is_empty"]

    if is_empty:
        print("[orange]Dataset file is not registered to the version[/orange]")
        return

    # storage_path = dataset_details["path"]["source_path"]
    # storage_source_type = dataset_details["path"]["source_type"]["public_url"]

    # dataset_url = urljoin(storage_source_type, storage_path)

    dataset_url = dataset_details["path"]
    source_type = dataset_details["source_type"]

    if source_type != "file":
        response = requests.get(dataset_url)

        if response.ok:
            dataset_bytes = response.content
            # open('temp_dataset.parquet', 'wb').write(dataset_bytes)
            # dataset = pd.read_parquet('temp_dataset.parquet')

            open("temp_dataset.pkl", "wb").write(dataset_bytes)
            dataset = joblib.load("temp_dataset.pkl")

            # print(f"[green]Dataset has been fetched[/green]")
            return dataset
        else:
            print(f"[orange]Unable to fetch Dataset[/orange]")
            # print(response.status_code)
            # print(response.text)
            # print(response.url)
            return
    else:
        if os.path.exists(dataset_url):
            dataset = joblib.load(dataset_url)
            return dataset
        else:
            print(f"[orange]Unable to fetch Dataset[/orange]")
            return


# def delete(label: str) -> str:
#     """This function deletes a dataset from the project

#     Parameters
#     ----------
#     name : str
#         The name of the dataset you want to delete
#     version : str
#         The version of the dataset to delete.

#     """

#     name, version = parse_version_label(label)

#     user_token = get_token()
#     org_id = get_org_id()
#     dataset_schema = DatasetSchema()

#     url = "org/{}/dataset/{}/delete".format(org_id, name)
#     url = urljoin(dataset_schema.backend.BASE_URL, url)

#     headers = {
#         "Content-Type": ContentTypeHeader.APP_FORM_URL_ENCODED.value,
#         "Authorization": "Bearer {}".format(user_token),
#     }

#     data = {"version": version}

#     data = json.dumps(data)

#     response = requests.delete(url, headers=headers, data=data)

#     if response.ok:
#         print(f"[green]Dataset has been deleted[/green]")

#     else:
#         print(f"[orange]Unable to delete Dataset[/orange]")

#     return response.text




def merge_step_with_value(values_dict, step):
    value_dict_with_step = {}

    value_dict_with_step[step] = values_dict

    
    return value_dict_with_step



def update_step_dict(values_dict_existing, values_dict_new):

    for step_new in values_dict_new.keys():

        if step_new not in values_dict_existing.keys():
            values_dict_existing[step_new] = values_dict_new[step_new]
        else:
            values_dict_existing[step_new].update(values_dict_new[step_new])

    return values_dict_existing



BASE_URL = "https://pureml-development.up.railway.app/api"


# DataSet

DatasetCreate = BASE_URL + "/org/{orgId}/dataset/{datasetName}/create"  #  Create dataset
DatasetRegister = BASE_URL  + "/org/{orgId}/dataset/{datasetName}/register" #Register a dataset


# Model

ModelCreate = BASE_URL + "/org/{orgId}/model/{modelName}/create"   # Create Model
ModelRisk = BASE_URL + "/org/{orgId}/model/{modelName}/version/{version}/risk" # Risk Data for a Model
ModelMonitor = BASE_URL + "/org/{orgId}/model/{modelName}/version/{version}/monitor" # Monitoring data for a model
ModelRegister = BASE_URL + "/org/{orgId}/model/{modelName}/register" # Register a Model




from pureml.components.model import register
from pureml.utils.pipeline import (
    add_model_to_config,
    load_metrics_from_config,
    load_params_from_config,
    load_figures_from_config,
)
from pureml import metrics, params, figure
from pureml.utils.version_utils import parse_version_label, generate_label
import functools
from pureml.utils.config import reset_config
from pureml.schema import ConfigKeys


config_keys = ConfigKeys


def model(label: str):
    def decorator(func):
        # print('Inside decorator')
        name, _ = parse_version_label(label)

        # print('Adding model name: ', name, 'to config before invoking user function')
        add_model_to_config(name=name)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # print("Inside wrapper")
            func_name = func.__name__
            func_description = func.__doc__

            func_output = func(*args, **kwargs)

            model_exists_in_remote, model_hash, model_version = register(
                model=func_output, label=label
            )

            if (
                model_exists_in_remote
            ):  # Only add the model to config if it is successfully pushed

                label_new = generate_label(name, model_version)

                add_model_to_config(
                    name=name,
                    description=func_description,
                    hash=model_hash,
                    version=model_version,
                    func=func,
                )

                metric_values = load_metrics_from_config()
                if len(metric_values) != 0:
                    metrics.add(metrics=metric_values, label=label_new)
                    reset_config(key=config_keys.metrics.value)

                param_values = load_params_from_config()
                if len(param_values) != 0:
                    params.add(params=param_values, label=label_new)
                    reset_config(key=config_keys.params.value)

                figure_file_paths = load_figures_from_config()
                if len(figure_file_paths) != 0:
                    figure.add(file_paths=figure_file_paths, label=label_new)
                    reset_config(key=config_keys.figure.value)

            else:
                add_model_to_config(
                    name=name,
                    description=func_description,
                    hash=model_hash,
                    func=func,
                )

            return func_output

        # print("Outside  wrapper")

        return wrapper

    # print('Outside decorator')

    return decorator



from enum import Enum


class ConfigKeys(Enum):
    load_data = "load_data"
    transformer = "transformer"
    dataset = "dataset"
    model = "model"
    params = "params"
    metrics = "metrics"
    figure = "figure"
    artifacts = "artifacts"
    pred_function = "pred_function"
    pip_requirement = "pip_requirement"
    resource = "resource"



from pureml.components.dataset import register
from pureml.utils.pipeline import add_dataset_to_config
from pureml.lineage.data.create_lineage import create_lineage
from pureml.utils.version_utils import parse_version_label
import functools
from pureml.utils.config import reset_config
from pureml.schema import ConfigKeys

config_keys = ConfigKeys


def dataset(label: str, parent: str = None, upload=False):
    def decorator(func):
        name, _ = parse_version_label(label)

        # Add dataset name to config here if it is being used by any of the pipeline components.
        # add_dataset_to_config(name=name, parent=parent)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            func_name = func.__name__
            func_description = func.__doc__

            func_output = func(*args, **kwargs)

            is_empty = False

            if not upload or func_output is None:
                is_empty = True

            add_dataset_to_config(
                name=name,
                description=func_description,
                parent=parent,
                func=func,
            )

            lineage = create_lineage()

            dataset_exists_in_remote, dataset_hash, dataset_version = register(
                dataset=func_output,
                label=label,
                lineage=lineage,
                is_empty=is_empty,
            )

            # Uncomment this if there any components that depend on dataset version, or dataset hash
            # if dataset_exists_in_remote:
            #     add_dataset_to_config(name=name, hash=dataset_hash, version=dataset_version, parent=parent, func=func)
            add_dataset_to_config(
                name=name,
                description=func_description,
                hash=dataset_hash,
                parent=parent,
                func=func,
            )

            reset_config(key=config_keys.load_data.value)
            reset_config(key=config_keys.transformer.value)
            reset_config(key=config_keys.dataset.value)

            return func_output

        return wrapper

    return decorator



import structlog
import logging
import os
from datetime import datetime
from inspect import getframeinfo, stack


class CallerFileProcessor(object):
    def __init__(self, stack_depth) -> None:
        self.stack_depth = stack_depth

    def __call__(self, _, __, event_dict):
        caller = getframeinfo(stack()[self.stack_depth][0])
        event_dict["file"] = (
            f"{caller.filename.replace(os.path.abspath('../../..'), '')[1:]}:{caller.lineno}"
        )
        return event_dict


def __get_fmt_time() -> str:
    return datetime.now().strftime("%d-%m-%Y_%H%M%S")


structlog.configure(
    wrapper_class=structlog.make_filtering_bound_logger(logging.INFO),
    processors=[
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.processors.add_log_level,
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        CallerFileProcessor(stack_depth=4),
        structlog.processors.KeyValueRenderer(),
    ],
    logger_factory=structlog.WriteLoggerFactory(
        file=open(
            os.path.join(os.path.expanduser("~"), f"pureml_{__get_fmt_time()}.log"),
            "wt",
        ),
    ),
)


def get_logger(name: str):
    return structlog.get_logger(name=name)



import requests

from rich import print
from rich.syntax import Syntax

import os
import json


from urllib.parse import urljoin

from pureml.cli.helpers import get_auth_headers

from . import get_token, get_org_id

from pureml.schema import PathSchema, BackendSchema, ContentTypeHeader
from joblib import Parallel, delayed
from PIL import Image

#path_schema = PathSchema().get_instance()
#backend_schema = BackendSchema().get_instance()
path_schema = PathSchema()
backend_schema = BackendSchema()

def save_images(array):
    array_paths = {}
    for array_key, array_value in array.items():
        save_name = os.path.join(
            path_schema.PATH_ARRAY_DIR, ".".join(array_key, ".png")
        )

        data = Image.fromarray(array_value)
        data.save(save_name)

        array_paths[array_key] = array_paths

    return array_paths


def details(
    model_name: str, model_version: str = "latest", name: str = ""
):
    """This function returns the details of the array for a given model

    Parameters
    ----------
    model_name : str
        The name of the model you want to get the array details for
    model_version: str
        The version of the model
    name : str
        The name of the array.

    """

    org_id = get_org_id()

    url = "org/{}/model/{}/version/{}/log".format(
        org_id, model_name, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.status_code == 200:
        res_text = json.loads(response.text)

        if len(res_text) == 0:
            print("[yellow] No array have been found for the model")
            print(res_text)
            return
        else:
            print("[green]array have been found for the model")
            print(res_text)
            return res_text

    else:
        print("[orange]Unable to obtain the array details")
        print(response.text)
        return


def add(
    array: str, model_name: str, model_version: str = "latest"
) -> str:
    """`add` function takes in the path of the array, name of the array and the model name and
    registers the array

    Parameters
    ----------
    array : str
        The path to the array file.
    name : str
        The name of the array.
    model_name : str
        The name of the model you want to add array to.
    model_version: str
        The version of the model

    Returns
    -------
        The response is a JSON object

    """

    org_id = get_org_id()

    url = "org/{}/model/{}/version/{}/log".format(
        org_id, model_name, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    array_paths = save_images(arrays=array)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    files = {}
    for file_name, file_path in array_paths.items():

        if os.path.isfile(file_path):
            files[file_name] = open(file_path, "rb")
        else:
            print("[orange] array", file_name, "does not exist at the given path")

    data = {"name_path_mapping": array_paths}

    response = requests.post(url, data=data, files=files, headers=headers)

    if response.status_code == 200:
        print(f"[green]arrays have been registered!")

    else:
        print(f"[orange]arrays have not been registered!")
        print(response.text)

    return response.text


def fetch(
    model_name: str, model_version: str = "latest", name: str = ""
):
    """It fetches the array from the server and stores it in the local directory

    Parameters
    ----------
    model_name : str
        The name of the model you want to fetch the array from.
    model_version: str
        The version of the model
    name : str
        The name of the array to be fetched. If not specified, all arrays will be fetched.

    Returns
    -------
        The response text is being returned.

    """

    org_id = get_org_id()

    def fetch_array(array_details: dict):

        url = array_details["location"]
        file_path_temp = array_details["path"]
        file_name = file_path_temp.split(os.path.sep)[-1]
        save_path = os.path.join(path_schema.PATH_ARRAY_DIR, file_name)
        print("save path", save_path)

        name_fetched = array_details["array"]

        headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

        print("array url", url)

        # response = requests.get(url, headers=headers)
        response = requests.get(url)

        print(response.status_code)

        if response.status_code == 200:
            print("[green] array {} has been fetched".format(name_fetched))

            save_dir = os.path.dirname(save_path)

            os.makedirs(save_dir, exist_ok=True)

            array_bytes = response.content

            open(save_path, "wb").write(array_bytes)

            print(
                "[green] array {} has been stored at {}".format(
                    name_fetched, save_path
                )
            )

            return response.text
        else:
            print("[orange] Unable to fetch the array")

            return response.text

    array_details = details(
        model_name=model_name,
        name=name,
        model_version=model_version,
    )

    if array_details is None:
        return

    if type(array_details) is dict:

        res_text = fetch_array(array_details)

    elif type(array_details) is list:
        res_text = Parallel(n_jobs=-1)(
            delayed(fetch_array)(art_det) for art_det in array_details
        )

    return res_text


def delete(
    name: str, model_name: str, model_version: str = "latest"
) -> str:
    """`delete()` deletes an array from a model

    Parameters
    ----------
    name : str
        The name of the array you want to delete.
    model_name : str
        The name of the model you want to delete the array from
    model_version: str
        The version of the model

    """

    org_id = get_org_id()

    url = "org/{}/model/{}/version/{}/log".format(
        org_id, model_name, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    # array_details = details(model_name=model_name, array=array)

    # if array_details is None:
    #     print('[orange] Unable to find array details')
    #     return

    response = requests.delete(url, headers=headers)

    if response.status_code == 200:
        print(f"[green]array has been deleted")

    else:
        print(f"[orange]Unable to delete array")

    return response.text



import importlib.metadata as importlib_metadata
import sys
import gc





def get_file_size(input_obj):
    memory_size = 0
    ids = set()
    objects = [input_obj]
    while objects:
        new = []
        for obj in objects:
            if id(obj) not in ids:
                ids.add(id(obj))
                memory_size += sys.getsizeof(obj)
                new.append(obj)
        objects = gc.get_referents(*new)
    return memory_size




def get_framework_version(framework_name: str) -> str:
    try:
        get_pkg_version = importlib_metadata.version

        framework_version = get_pkg_version(framework_name)


    except importlib_metadata.PackageNotFoundError:
    # Note `importlib_metadata.version(package)` is not necessarily equal to
    # `__import__(package).__version__`. See the example for pytorch below.
    #
    # Example
    # -------
    # $ pip install torch==1.9.0
    # $ python -c "import torch; print(torch.__version__)"
    # 1.9.0+cu102
    # $ python -c "import importlib_metadata; print(importlib_metadata.version('torch'))"
    # 1.9.0
        framework_version = __import__(framework_name).__version__

    return framework_version



def infer_requirements(framework_name) -> str:
    framework_version = get_framework_version(framework_name)

    pip_requirements = f"{framework_name}=={framework_version}"

    return pip_requirements


from pydantic import BaseModel
from .paths import PathSchema
from .backend import BackendSchema
import os
from enum import Enum


class LogKeys(Enum):
    metrics = "metrics"
    params = "params"
    figure = "figure"
    predict = "predict"
    requirements = "requirements"
    resources = "resources"


class LogSchema(BaseModel):
    _paths: PathSchema = PathSchema().get_instance()
    _backend: BackendSchema = BackendSchema().get_instance()
    key: LogKeys = LogKeys.metrics



from pathlib import Path
from typing import Optional
import jwt
import requests

# import typer
from rich import print
from rich.syntax import Syntax

import os
import json
import typing

from urllib.parse import urljoin

from pureml.cli.helpers import get_auth_headers

from . import get_org_id

from pureml.schema import PathSchema, BackendSchema, ContentTypeHeader
from joblib import Parallel, delayed


path_schema = PathSchema()
backend_schema = BackendSchema()
#path_schema = PathSchema().get_instance()
#backend_schema = BackendSchema().get_instance()


def add(
    video: str, model_name: str, model_version: str = "latest"
) -> str:
    """`add` function takes in the path of the video, name of the video and the model name and
    registers the video

    Parameters
    ----------
    video : str
        The path to the video file.
    name : str
        The name of the video.
    model_name : str
        The name of the model you want to add video to.
    model_version: str
        The version of the model

    Returns
    -------
        The response is a JSON object

    """

    org_id = get_org_id()

    url = "org/{}/model/{}/version/{}/log".format(
        org_id, model_name, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    files = {}
    for file_name, file_path in video.items():

        if os.path.isfile(file_path):
            files[file_name] = open(file_path, "rb")
        else:
            print("[orange] video", file_name, "does not exist at the given path")

    data = {"name_path_mapping": video}

    response = requests.post(url, data=data, files=files, headers=headers)

    if response.status_code == 200:
        print(f"[green]videos have been registered!")

    else:
        print(f"[orange]videos have not been registered!")
        print(response.text)

    return response.text


def details(model_name: str, model_version: str = "latest"):
    org_id = get_org_id()

    url = "org/{}/model/{}/version/{}/log".format(
        org_id, model_name, model_version
    )
    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

    response = requests.get(url, headers=headers)

    if response.ok:
        # T-1161 standardize api response to contains Models as a list
        response_text = response.json()
        details = response_text["data"]

        # print(model_details)

        return details

    else:
        print(f"[orange]Model details details have not been found")
        return


def fetch(
    model_name: str, model_version: str = "latest", name: str = ""
):
    """It fetches the video from the server and stores it in the local directory

    Parameters
    ----------
    model_name : str
        The name of the model you want to fetch the video from.
    model_version: str
        The version of the model
    name : str
        The name of the video to be fetched. If not specified, all videos will be fetched.

    Returns
    -------
        The response text is being returned.

    """

    org_id = get_org_id()

    def fetch_video(video_details: dict):

        url = video_details["location"]
        file_path_temp = video_details["path"]
        file_name = file_path_temp.split(os.path.sep)[-1]
        save_path = os.path.join(path_schema.PATH_VIDEO_DIR, file_name)
        print("save path", save_path)

        name_fetched = video_details["video"]

        headers = get_auth_headers(content_type=ContentTypeHeader.APP_FORM_URL_ENCODED)

        print("video url", url)

        # response = requests.get(url, headers=headers)
        response = requests.get(url)

        print(response.status_code)

        if response.status_code == 200:
            print("[green] video {} has been fetched".format(name_fetched))

            save_dir = os.path.dirname(save_path)

            os.makedirs(save_dir, exist_ok=True)

            video_bytes = response.content

            open(save_path, "wb").write(video_bytes)

            print(
                "[green] video {} has been stored at {}".format(
                    name_fetched, save_path
                )
            )

            return response.text
        else:
            print("[orange] Unable to fetch the video")

            return response.text

    video_details = details(
        model_name=model_name,
        name=name,
        model_version=model_version,
    )

    if video_details is None:
        return

    if type(video_details) is dict:

        res_text = fetch_video(video_details)

    elif type(video_details) is list:
        res_text = Parallel(n_jobs=-1)(
            delayed(fetch_video)(art_det) for art_det in video_details
        )

    return res_text


# def delete(name:str, model_name:str,  model_version:str='latest') -> str:
#     '''`delete()` deletes an video from a model

#     Parameters
#     ----------
#     name : str
#         The name of the video you want to delete.
#     model_name : str
#         The name of the model you want to delete the video from
#     model_version: str
#         The version of the model

#     '''

#     user_token = get_token()
#     org_id = get_org_id()
#     project_id = get_project_id()

#     url_path_1 = '{}/project/{}/model/{}/{}/video/{}/delete'.format(org_id, project_id, model_name, model_version, name)
#     url = urljoin(BASE_URL, url_path_1)


#     headers = {
#         'Content-Type': 'application/x-www-form-urlencoded',
#         'Authorization': 'Bearer {}'.format(user_token)
#     }


#     # video_details = details(model_name=model_name, video=video)

#     # if video_details is None:
#     #     print('[orange] Unable to find video details')
#     #     return


#     response = requests.delete(url, headers=headers)


#     if response.status_code == 200:
#         print(f"[green]video has been deleted")

#     else:
#         print(f"[orange]Unable to delete video")

#     return response.text



import boto3
from ..provider import StorageProvider
from pureml.utils.logger import get_logger

logger = get_logger("sdk.storage.cloudflare")


class R2StorageProvider(StorageProvider):
    def __init__(
        self,
        access_key_id: str,
        access_key_secret: str,
        bucket_name: str,
        account_id: str,
        public_url: str,
        bucket_location: str = None,
    ):
        self.r2 = boto3.client(
            "s3",
            aws_access_key_id=access_key_id,
            aws_secret_access_key=access_key_secret,
            endpoint_url=f"https://{account_id}.r2.cloudflarestorage.com",
        )
        self.bucket_name = bucket_name
        if public_url:
            self.public_url = public_url
        else:
            print("Public URL not found in Cloudflare secrets")

    def upload_file(self, source_file, destination) -> str or None:
        destination = destination.lstrip("/")
        try:
            self.r2.upload_file(source_file, self.bucket_name, destination)
            return self.public_url + "/" + destination
        except Exception as e:
            logger.error("Error uploading file to R2", error=e)
            print(e)
            return None

    def download_file(self, source, destination_file):
        source = source.lstrip("/")
        try:
            self.r2.download_file(self.bucket_name, source, destination_file)
        except Exception as e:
            logger.error("Error downloading file from R2", error=e)
            print(e)
